import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { MagnetWheel } from './MagnetWheel.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
const ANGLE = 0;
const ANGLEP = 1;
const TIME = 2;
const KE = 3;
const PE = 4;
const TE = 5;
export class MagnetWheelSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.damping_ = 0.7;
        this.potentialOffset_ = 0;
        this.isDragging = false;
        this.keyLeft_ = false;
        this.keyRight_ = false;
        this.keyForce_ = 5;
        const var_names = [
            MagnetWheelSim.en.ANGLE,
            MagnetWheelSim.en.ANGULAR_VELOCITY,
            VarsList.en.TIME,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY
        ];
        const i18n_names = [
            MagnetWheelSim.i18n.ANGLE,
            MagnetWheelSim.i18n.ANGULAR_VELOCITY,
            VarsList.i18n.TIME,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.wheel_ = new MagnetWheel(1, 'wheel');
        this.saveInitialState();
        this.getSimList().add(this.wheel_);
        this.addParameter(new ParameterNumber(this, MagnetWheelSim.en.DAMPING, MagnetWheelSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, MagnetWheelSim.en.MASS, MagnetWheelSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, MagnetWheelSim.en.MAGNET_STRENGTH, MagnetWheelSim.i18n.MAGNET_STRENGTH, () => this.getMagnetStrength(), a => this.setMagnetStrength(a))
            .setLowerLimit(Number.NEGATIVE_INFINITY));
        this.addParameter(new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a))
            .setLowerLimit(Number.NEGATIVE_INFINITY)
            .setSignifDigits(5));
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', wheel_: ' + this.wheel_
            + ', damping_: ' + Util.NF(this.damping_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'MagnetWheelSim';
    }
    ;
    getEnergyInfo() {
        const re = this.wheel_.rotationalEnergy();
        const pe = this.wheel_.getPotentialEnergy() + this.potentialOffset_;
        return new EnergyInfo(pe, 0, re);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        vars[KE] = ei.getRotational();
        vars[PE] = ei.getPotential();
        vars[TE] = ei.getTotalEnergy();
        va.setValues(vars, true);
    }
    ;
    moveObjects(vars) {
        this.wheel_.setAngle(vars[ANGLE]);
        this.wheel_.setAngularVelocity(vars[ANGLEP]);
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.wheel_) {
            this.isDragging = true;
            return true;
        }
        return false;
    }
    ;
    mouseDrag(_simObject, _location, _offset) {
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging = false;
    }
    ;
    handleKeyEvent(evt, pressed, modifiers) {
        if (modifiers.control || modifiers.meta || modifiers.alt)
            return;
        switch (evt.key) {
            case "ArrowLeft":
            case "J":
            case "j":
                this.keyLeft_ = pressed;
                evt.preventDefault();
                break;
            case "ArrowRight":
            case "L":
            case "l":
                this.keyRight_ = pressed;
                evt.preventDefault();
                break;
            default:
                break;
        }
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        const m = this.wheel_.getMass();
        change[ANGLE] = vars[ANGLEP];
        change[ANGLEP] = -this.damping_ * vars[ANGLEP] / m;
        change[TIME] = 1.0;
        this.moveObjects(vars);
        const fm = this.wheel_.getFixedMagnet();
        const magnets = this.wheel_.getMagnets();
        this.wheel_.calculateForces().forEach(f => {
            change[ANGLEP] += f.getTorque() / m;
            f.setExpireTime(this.getTime());
            this.getSimList().add(f);
        });
        if (this.keyLeft_) {
            change[ANGLEP] += this.keyForce_;
        }
        else if (this.keyRight_) {
            change[ANGLEP] += -this.keyForce_;
        }
        return null;
    }
    ;
    getMass() {
        return this.wheel_.getMass();
    }
    ;
    setMass(value) {
        this.wheel_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(MagnetWheelSim.en.MASS);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(MagnetWheelSim.en.DAMPING);
    }
    ;
    getMagnetStrength() {
        return this.wheel_.getMagnetStrength();
    }
    ;
    setMagnetStrength(value) {
        this.wheel_.setMagnetStrength(value);
        this.broadcastParameter(MagnetWheelSim.en.MAGNET_STRENGTH);
    }
    ;
    getMagnetWheel() {
        return this.wheel_;
    }
    ;
}
MagnetWheelSim.en = {
    DAMPING: 'damping',
    MASS: 'mass',
    ANGLE: 'angle',
    ANGULAR_VELOCITY: 'angular velocity',
    MAGNET_STRENGTH: 'magnet strength',
    NUM_MAGNETS: 'number of magnets',
    SYMMETRIC: 'symmetric',
    MAGNET_ANGLE: 'angle between magnets'
};
MagnetWheelSim.de_strings = {
    DAMPING: 'Dämpfung',
    MASS: 'Masse',
    ANGLE: 'Winkel',
    ANGULAR_VELOCITY: 'Winkel Geschwindigkeit',
    MAGNET_STRENGTH: 'Magnet Stärke',
    NUM_MAGNETS: 'Anzahl Magnete',
    SYMMETRIC: 'symmetrisch',
    MAGNET_ANGLE: 'Winkel zwischen Magnete'
};
MagnetWheelSim.i18n = Util.LOCALE === 'de' ? MagnetWheelSim.de_strings : MagnetWheelSim.en;
Util.defineGlobal('sims$misc$MagnetWheelSim', MagnetWheelSim);
