import { AbstractODESim } from '../../lab/model/ODESim.js';
import { AffineTransform } from '../../lab/util/AffineTransform.js';
import { Force } from '../../lab/model/Force.js';
import { GenericObserver, ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const X = 0;
const V = 1;
const TIME = 2;
const RPM = 3;
const ENGINE = 4;
const GRAVITY = 5;
export class RobotSpeedSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.radius_ = 0.045;
        this.torque_ = 1.98;
        this.freeSpeed_ = 317;
        this.mass_ = 2.5;
        this.slope_ = 0;
        this.limitTorque_ = true;
        this.start_ = Vector.ORIGIN;
        this.engineForce_ = 0;
        this.gravityRampForce_ = 0;
        this.gravityForce_ = 0;
        this.robot_ = PointMass.makeRectangle(0.3, 0.1, 'robot');
        this.cm_pct_ = 50;
        this.wheelf_ = PointMass.makeCircle(2 * this.radius_, 'wheelf');
        this.wheelr_ = PointMass.makeCircle(2 * this.radius_, 'wheelr');
        this.ramp_ = PointMass.makeRectangle(7, 0.025, 'ramp');
        this.vwf_ = Vector.ORIGIN;
        this.vwr_ = Vector.ORIGIN;
        this.Nf_ = 0;
        this.Nr_ = 0;
        this.friction_ = 1;
        const var_names = [
            RobotSpeedSim.en.POSITION,
            RobotSpeedSim.en.VELOCITY,
            VarsList.en.TIME,
            RobotSpeedSim.en.RPM,
            RobotSpeedSim.en.ENGINE_FORCE,
            RobotSpeedSim.en.GRAVITY_FORCE
        ];
        const i18n_names = [
            RobotSpeedSim.i18n.POSITION,
            RobotSpeedSim.i18n.VELOCITY,
            VarsList.i18n.TIME,
            RobotSpeedSim.i18n.RPM,
            RobotSpeedSim.i18n.ENGINE_FORCE,
            RobotSpeedSim.i18n.GRAVITY_FORCE
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(RPM, ENGINE, GRAVITY);
        this.getSimList().add(this.ramp_, this.robot_, this.wheelf_, this.wheelr_);
        let pn;
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.MASS, RobotSpeedSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        pn.setUnits(' (kg)');
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.DIAMETER, RobotSpeedSim.i18n.DIAMETER, () => this.getDiameter(), a => this.setDiameter(a)));
        pn.setUnits(' (mm)');
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.TORQUE, RobotSpeedSim.i18n.TORQUE, () => this.getTorque(), a => this.setTorque(a)));
        pn.setUnits(' (Nm)');
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.FREE_SPEED, RobotSpeedSim.i18n.FREE_SPEED, () => this.getFreeSpeed(), a => this.setFreeSpeed(a)));
        pn.setUnits(' (RPM)');
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.SLOPE, RobotSpeedSim.i18n.SLOPE, () => this.getSlope(), a => this.setSlope(a)));
        pn.setUnits(' (degrees)');
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.COEF_FRICTION, RobotSpeedSim.i18n.COEF_FRICTION, () => this.getFriction(), a => this.setFriction(a)));
        this.addParameter(pn = new ParameterNumber(this, RobotSpeedSim.en.CENTER_OF_MASS, RobotSpeedSim.i18n.CENTER_OF_MASS, () => this.getCenterOfMass(), a => this.setCenterOfMass(a)));
        pn.setUnits(' (%)');
        this.setCenterOfMass(50);
        new GenericObserver(this, (_e) => this.modifyObjects());
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', mass_: ' + Util.NF(this.mass_)
            + ', torque_: ' + Util.NF(this.torque_)
            + ', freeSpeed_: ' + Util.NF(this.freeSpeed_)
            + ', radius_: ' + Util.NF(this.radius_)
            + ', slope_: ' + Util.NF(this.slope_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'RobotSpeedSim';
    }
    ;
    reset() {
        this.Nf_ = 0;
        this.Nr_ = 0;
        this.engineForce_ = 0;
        this.gravityRampForce_ = 0;
        this.gravityForce_ = 0;
        super.reset();
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        const p = this.map_p_to_vector(vars[X]);
        const cs = Math.cos(this.slope_);
        const ss = Math.sin(this.slope_);
        const at = new AffineTransform(cs, ss, -ss, cs, p.getX(), p.getY());
        this.robot_.setPosition(p);
        this.robot_.setAngle(this.slope_);
        this.wheelf_.setPosition(at.transform(this.vwf_));
        this.wheelr_.setPosition(at.transform(this.vwr_));
        const n = new Vector(-ss, cs);
        let f = new Force('normal_f', this.wheelf_, this.wheelf_.getPosition(), 1, n.multiply(0.1 * this.Nf_), 1);
        f.setExpireTime(this.getTime());
        this.getSimList().add(f);
        f = new Force('normal_r', this.wheelr_, this.wheelr_.getPosition(), 1, n.multiply(0.1 * this.Nr_), 1);
        f.setExpireTime(this.getTime());
        this.getSimList().add(f);
        f = new Force('gravity', this.robot_, this.robot_.getPosition(), 1, Vector.NORTH.multiply(0.1 * this.gravityForce_), 1);
        f.setExpireTime(this.getTime());
        this.getSimList().add(f);
        f = new Force('engine', this.robot_, this.wheelr_.getPosition(), 1, new Vector(cs, ss).multiply(0.1 * this.engineForce_), 1);
        f.setExpireTime(this.getTime());
        this.getSimList().add(f);
        let a = -vars[X] / (2 * Math.PI * this.radius_);
        a = a - Math.floor(a);
        a = 2 * Math.PI * a;
        this.wheelf_.setAngle(a + this.slope_);
        this.wheelr_.setAngle(a + this.slope_);
        const at2 = new AffineTransform(cs, ss, -ss, cs, 0, 0);
        this.ramp_.setPosition(at2.transform(2.5, -0.075 - this.radius_
            - 0.5 * this.ramp_.getHeight()));
        this.ramp_.setAngle(this.slope_);
        vars[RPM] = vars[V] * 60 / (2 * Math.PI * this.radius_);
        vars[ENGINE] = this.engineForce_;
        vars[GRAVITY] = -this.gravityRampForce_;
        va.setValues(vars, true);
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[X] = vars[V];
        this.Nr_ = this.mass_ * 9.81 * Math.cos(this.slope_) /
            (1 - this.vwr_.getX() / this.vwf_.getX());
        this.Nf_ = -(this.vwr_.getX() / this.vwf_.getX()) * this.Nr_;
        const rpm = vars[V] * 60 / (2 * Math.PI * this.radius_);
        let t = (-this.torque_ / this.freeSpeed_) * rpm + this.torque_;
        if (rpm < 0) {
            t = this.torque_;
        }
        else if (rpm > this.freeSpeed_) {
            t = 0;
        }
        let f = t / this.radius_;
        if (this.limitTorque_) {
            const limit = this.friction_ * this.Nr_;
            if (f > limit) {
                f = limit;
            }
        }
        this.engineForce_ = f;
        this.gravityForce_ = -this.mass_ * 9.81;
        const g = this.gravityForce_ * Math.sin(this.slope_);
        this.gravityRampForce_ = g;
        change[V] = (f + g) / this.mass_;
        change[TIME] = 1.0;
        return null;
    }
    ;
    map_p_to_vector(p) {
        const s = this.slope_;
        return this.start_.add(new Vector(p * Math.cos(s), p * Math.sin(s)));
    }
    ;
    getDiameter() {
        return 2 * this.radius_ * 1000;
    }
    ;
    setDiameter(value) {
        this.radius_ = value / 2000;
        this.wheelf_.setWidth(2 * this.radius_);
        this.wheelf_.setHeight(2 * this.radius_);
        this.wheelr_.setWidth(2 * this.radius_);
        this.wheelr_.setHeight(2 * this.radius_);
        this.broadcastParameter(RobotSpeedSim.en.DIAMETER);
    }
    ;
    getMass() {
        return this.mass_;
    }
    ;
    setMass(value) {
        this.mass_ = value;
        this.broadcastParameter(RobotSpeedSim.en.MASS);
    }
    ;
    getSlope() {
        return 180 * this.slope_ / Math.PI;
    }
    ;
    setSlope(value) {
        this.slope_ = Math.PI * value / 180;
        this.broadcastParameter(RobotSpeedSim.en.SLOPE);
    }
    ;
    getTorque() {
        return this.torque_;
    }
    ;
    setTorque(value) {
        this.torque_ = value;
        this.broadcastParameter(RobotSpeedSim.en.TORQUE);
    }
    ;
    getFreeSpeed() {
        return this.freeSpeed_;
    }
    ;
    setFreeSpeed(value) {
        this.freeSpeed_ = value;
        this.broadcastParameter(RobotSpeedSim.en.FREE_SPEED);
    }
    ;
    getFriction() {
        return this.friction_;
    }
    ;
    setFriction(value) {
        this.friction_ = value;
        this.broadcastParameter(RobotSpeedSim.en.COEF_FRICTION);
    }
    ;
    getCenterOfMass() {
        return this.cm_pct_;
    }
    ;
    setCenterOfMass(value) {
        if (value < 1 || value > 99) {
            throw 'center of gravity percentage must be > 1 and < 99, was ' + value;
        }
        this.cm_pct_ = value;
        this.robot_.setCenterOfMass(new Vector(0.125 - (value / 100) * 0.25, 0));
        const cm = this.robot_.getCenterOfMass();
        this.vwf_ = new Vector(0.125, -0.075).subtract(cm);
        this.vwr_ = new Vector(-0.125, -0.075).subtract(cm);
        this.broadcastParameter(RobotSpeedSim.en.CENTER_OF_MASS);
    }
    ;
}
RobotSpeedSim.en = {
    POSITION: 'position',
    VELOCITY: 'velocity',
    RPM: 'RPM',
    TORQUE: 'stall torque',
    FREE_SPEED: 'free speed',
    MASS: 'mass',
    ROBOT: 'robot',
    SLOPE: 'slope',
    DIAMETER: 'wheel diameter',
    ENGINE_FORCE: 'engine force',
    GRAVITY_FORCE: 'gravity force',
    COEF_FRICTION: 'coef static friction',
    CENTER_OF_MASS: 'center of mass'
};
RobotSpeedSim.de_strings = {
    POSITION: 'Position',
    VELOCITY: 'Geschwindigkeit',
    RPM: 'RPM',
    TORQUE: 'stehenbleibenes Drehmoment',
    FREE_SPEED: 'freie Geschwindigkeit',
    MASS: 'Masse',
    ROBOT: 'Roboter',
    SLOPE: 'Neigung',
    DIAMETER: 'Rad Durchmesser',
    ENGINE_FORCE: 'Triebkraft',
    GRAVITY_FORCE: 'Schwerkraft',
    COEF_FRICTION: 'Koeff Statisch Reibung',
    CENTER_OF_MASS: 'Zentrum der Masse'
};
RobotSpeedSim.i18n = Util.LOCALE === 'de' ? RobotSpeedSim.de_strings : RobotSpeedSim.en;
Util.defineGlobal('sims$misc$RobotSpeedSim', RobotSpeedSim);
