import { AbstractODESim } from '../../lab/model/ODESim.js';
import { ConcreteLine } from '../../lab/model/ConcreteLine.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const X = 0;
const H = 1;
const XP = 2;
const HP = 3;
const WORK = 4;
const KE = 5;
const PE = 6;
const TE = 7;
const TIME = 8;
export class CartPendulumSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.isDragging_ = false;
        this.length_ = 1;
        this.gravity_ = 9.8;
        this.dampingCart_ = 0;
        this.dampingPendulum_ = 0;
        this.initialEnergy_ = 0;
        this.potentialOffset_ = 0;
        this.fixedPoint_ = PointMass.makeSquare(0.5, 'fixed point');
        this.rod_ = new ConcreteLine('rod');
        this.cart_ = PointMass.makeRectangle(1, 0.3, 'cart');
        this.pendulum_ = PointMass.makeCircle(0.3, 'bob');
        const var_names = [
            CartPendulumSim.en.CART_POSITION,
            CartPendulumSim.en.PENDULUM_ANGLE,
            CartPendulumSim.en.CART_VELOCITY,
            CartPendulumSim.en.PENDULUM_ANGLE_VELOCITY,
            CartPendulumSim.en.WORK_FROM_DAMPING,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME
        ];
        const i18n_names = [
            CartPendulumSim.i18n.CART_POSITION,
            CartPendulumSim.i18n.PENDULUM_ANGLE,
            CartPendulumSim.i18n.CART_VELOCITY,
            CartPendulumSim.i18n.PENDULUM_ANGLE_VELOCITY,
            CartPendulumSim.i18n.WORK_FROM_DAMPING,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME
        ];
        const va = new VarsList(var_names, i18n_names, this.getName() + '_VARS');
        this.setVarsList(va);
        va.setComputed(4, 5, 6, 7);
        this.fixedPoint_.setMass(Infinity);
        this.fixedPoint_.setPosition(new Vector(3, 0));
        this.spring_ = new Spring('spring', this.fixedPoint_, Vector.ORIGIN, this.cart_, Vector.ORIGIN, 3.0, 6.0);
        this.getSimList().add(this.fixedPoint_, this.rod_, this.cart_, this.pendulum_, this.spring_);
        this.initWork();
        this.saveInitialState();
        let pn;
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.GRAVITY, CartPendulumSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.CART_MASS, CartPendulumSim.i18n.CART_MASS, () => this.getCartMass(), a => this.setCartMass(a)));
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.PENDULUM_MASS, CartPendulumSim.i18n.PENDULUM_MASS, () => this.getPendulumMass(), a => this.setPendulumMass(a)));
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.CART_DAMPING, CartPendulumSim.i18n.CART_DAMPING, () => this.getCartDamping(), a => this.setCartDamping(a)));
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.PENDULUM_DAMPING, CartPendulumSim.i18n.PENDULUM_DAMPING, () => this.getPendulumDamping(), a => this.setPendulumDamping(a)));
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.PENDULUM_LENGTH, CartPendulumSim.i18n.PENDULUM_LENGTH, () => this.getPendulumLength(), a => this.setPendulumLength(a)));
        this.addParameter(new ParameterNumber(this, CartPendulumSim.en.SPRING_STIFFNESS, CartPendulumSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', dampingCart_: ' + Util.NF(this.dampingCart_)
            + ', dampingPendulum_: ' + Util.NF(this.dampingPendulum_)
            + ', length_: ' + Util.NF(this.length_)
            + ', cart_: ' + this.cart_
            + ', pendulum_: ' + this.pendulum_
            + ', spring_: ' + this.spring_
            + ', rod_: ' + this.rod_
            + ', fixedPoint_: ' + this.fixedPoint_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'CartPendulumSim';
    }
    ;
    initWork() {
        this.getVarsList().setValue(4, 0);
        this.initialEnergy_ = this.getEnergyInfo().getTotalEnergy();
    }
    ;
    getEnergyInfo() {
        const vars = this.getVarsList().getValues();
        return this.getEnergyInfo_(vars[WORK]);
    }
    ;
    getEnergyInfo_(work) {
        let ke = this.cart_.getKineticEnergy();
        ke += this.pendulum_.getKineticEnergy();
        let pe = this.spring_.getPotentialEnergy();
        const y = this.pendulum_.getPosition().getY();
        pe += this.gravity_ * this.pendulum_.getMass() * (y + this.length_);
        return new EnergyInfo(pe + this.potentialOffset_, ke, NaN, work, this.initialEnergy_);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        const angle = Util.limitAngle(vars[H]);
        if (angle != vars[H]) {
            this.getVarsList().setValue(1, angle);
            vars[H] = angle;
        }
        this.moveObjects(vars);
        const ei = this.getEnergyInfo_(vars[WORK]);
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        const angle = vars[H];
        const sinAngle = Math.sin(angle);
        const cosAngle = Math.cos(angle);
        this.cart_.setPosition(new Vector(vars[X], 0));
        this.cart_.setVelocity(new Vector(vars[XP], 0, 0));
        this.pendulum_.setPosition(new Vector(this.cart_.getPosition().getX() + this.length_ * sinAngle, this.cart_.getPosition().getY() - this.length_ * cosAngle));
        const vx = vars[XP] + this.length_ * vars[HP] * cosAngle;
        const vy = this.length_ * vars[HP] * sinAngle;
        this.pendulum_.setVelocity(new Vector(vx, vy));
        this.rod_.setStartPoint(this.cart_.getPosition());
        this.rod_.setEndPoint(this.pendulum_.getPosition());
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.cart_ || simObject == this.pendulum_) {
            this.isDragging_ = true;
            return true;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const vars = va.getValues();
        const p = location.subtract(offset);
        if (simObject == this.cart_) {
            vars[X] = p.getX();
            vars[XP] = 0;
            vars[HP] = 0;
        }
        else if (simObject == this.pendulum_) {
            const x1 = vars[X];
            const y1 = 0;
            const x2 = p.getX();
            const y2 = p.getY();
            const th = Math.atan2(x2 - x1, -(y2 - y1));
            vars[H] = th;
            vars[XP] = 0;
            vars[HP] = 0;
        }
        va.setValues(vars);
        this.initWork();
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging_ = false;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        this.moveObjects(vars);
        change[TIME] = 1;
        if (!this.isDragging_) {
            const m = this.pendulum_.getMass();
            const M = this.cart_.getMass();
            const L = this.length_;
            const k = this.spring_.getStiffness();
            const sh = Math.sin(vars[H]);
            const csh = Math.cos(vars[H]);
            change[X] = vars[XP];
            change[H] = vars[HP];
            let numer = m * vars[HP] * vars[HP] * L * sh + m * this.gravity_ * sh * csh - k * vars[X]
                - this.dampingCart_ * vars[XP] + this.dampingPendulum_ * vars[HP] * csh / L;
            change[XP] = numer / (M + m * sh * sh);
            numer = -m * vars[HP] * vars[HP] * L * sh * csh + k * vars[X] * csh - (M + m) * this.gravity_ * sh
                + this.dampingCart_ * vars[XP] * csh;
            numer += -(m + M) * this.dampingPendulum_ * vars[HP] / (m * L);
            change[HP] = numer / (L * (M + m * sh * sh));
            change[WORK] = -this.dampingCart_ * vars[XP] * vars[XP]
                - this.dampingPendulum_ * vars[HP] * vars[HP];
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.initWork();
        this.getVarsList().incrSequence(6, 7);
        this.broadcastParameter(CartPendulumSim.en.GRAVITY);
    }
    ;
    getPendulumMass() {
        return this.pendulum_.getMass();
    }
    ;
    setPendulumMass(value) {
        this.pendulum_.setMass(value);
        this.initWork();
        this.getVarsList().incrSequence(5, 6, 7);
        this.broadcastParameter(CartPendulumSim.en.PENDULUM_MASS);
    }
    ;
    getCartMass() {
        return this.cart_.getMass();
    }
    ;
    setCartMass(value) {
        this.cart_.setMass(value);
        this.initWork();
        this.getVarsList().incrSequence(5, 6, 7);
        this.broadcastParameter(CartPendulumSim.en.CART_MASS);
    }
    ;
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    ;
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.initWork();
        this.getVarsList().incrSequence(6, 7);
        this.broadcastParameter(CartPendulumSim.en.SPRING_STIFFNESS);
    }
    ;
    getCartDamping() {
        return this.dampingCart_;
    }
    ;
    setCartDamping(value) {
        this.dampingCart_ = value;
        this.initWork();
        this.broadcastParameter(CartPendulumSim.en.CART_DAMPING);
    }
    ;
    getPendulumDamping() {
        return this.dampingPendulum_;
    }
    ;
    setPendulumDamping(value) {
        this.dampingPendulum_ = value;
        this.initWork();
        this.broadcastParameter(CartPendulumSim.en.PENDULUM_DAMPING);
    }
    ;
    getPendulumLength() {
        return this.length_;
    }
    ;
    setPendulumLength(value) {
        this.length_ = value;
        this.initWork();
        this.getVarsList().incrSequence(5, 6, 7);
        this.broadcastParameter(CartPendulumSim.en.PENDULUM_LENGTH);
    }
    ;
}
CartPendulumSim.en = {
    CART_POSITION: 'cart position',
    PENDULUM_ANGLE: 'pendulum angle',
    CART_VELOCITY: 'cart velocity',
    PENDULUM_ANGLE_VELOCITY: 'pendulum angle velocity',
    CART_DAMPING: 'cart damping',
    PENDULUM_DAMPING: 'pendulum damping',
    GRAVITY: 'gravity',
    CART_MASS: 'cart mass',
    PENDULUM_MASS: 'pendulum mass',
    PENDULUM_LENGTH: 'pendulum length',
    SPRING_STIFFNESS: 'spring stiffness',
    WORK_FROM_DAMPING: 'work from damping'
};
CartPendulumSim.de_strings = {
    CART_POSITION: 'Wagenposition',
    PENDULUM_ANGLE: 'Pendelwinkel',
    CART_VELOCITY: 'Wagengeschwindigkeit',
    PENDULUM_ANGLE_VELOCITY: 'Pendelwinkelgeschwindigkeit',
    CART_DAMPING: 'Wagendämpfung',
    PENDULUM_DAMPING: 'Pendeldämpfung',
    GRAVITY: 'Gravitation',
    CART_MASS: 'Wagenmasse',
    PENDULUM_MASS: 'Pendelmasse',
    PENDULUM_LENGTH: 'Pendellänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    WORK_FROM_DAMPING: 'Arbeit von Dämpfung'
};
CartPendulumSim.i18n = Util.LOCALE === 'de' ? CartPendulumSim.de_strings : CartPendulumSim.en;
Util.defineGlobal('sims$pendulum$CartPendulumSim', CartPendulumSim);
