import { AbstractSubject } from '../../lab/util/AbstractSubject.js';
import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { Clock } from '../../lab/util/Clock.js';
import { CommonControls } from '../common/CommonControls.js';
import { CompareGraph } from '../common/CompareGraph.js';
import { CompareTimeGraph } from '../common/CompareTimeGraph.js';
import { DisplayArc } from '../../lab/view/DisplayArc.js';
import { DisplayClock } from '../../lab/view/DisplayClock.js';
import { DisplayLine } from '../../lab/view/DisplayLine.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { EnergyBarGraph } from '../../lab/graph/EnergyBarGraph.js';
import { GenericObserver } from '../../lab/util/Observe.js';
import { GraphLine } from '../../lab/graph/GraphLine.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PendulumSim } from './PendulumSim.js';
import { SimController } from '../../lab/app/SimController.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { SimRunner } from '../../lab/app/SimRunner.js';
import { SimView } from '../../lab/view/SimView.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { TabLayout } from '../common/TabLayout.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class ComparePendulumApp extends AbstractSubject {
    constructor(elem_ids) {
        Util.setErrorHandler();
        super('APP');
        this.angleDelta = 0.001;
        this.layout = new TabLayout(elem_ids);
        this.layout.getSimCanvas().setBackground('black');
        this.layout.getSimCanvas().setAlpha(CommonControls.SHORT_TRAILS);
        this.terminal = this.layout.getTerminal();
        const simCanvas = this.layout.getSimCanvas();
        this.sim1 = new PendulumSim('SIM_1');
        const startAngle = Math.PI / 4;
        const va1 = this.sim1.getVarsList();
        va1.setValue(0, startAngle);
        this.sim1.saveInitialState();
        this.simList = this.sim1.getSimList();
        new GenericObserver(this.sim1, (_evt) => this.sim1.modifyObjects(), 'modifyObjects after parameter or variable change');
        this.sim2 = new PendulumSim('SIM_2');
        this.terminal.setAfterEval(() => {
            this.sim1.modifyObjects();
            this.sim2.modifyObjects();
        });
        new GenericObserver(this.sim2, (_evt) => this.sim2.modifyObjects(), 'modifyObjects after parameter or variable change');
        const va2 = this.sim2.getVarsList();
        va2.setValue(0, startAngle + this.angleDelta);
        this.sim2.saveInitialState();
        this.simList2 = this.sim2.getSimList();
        this.simCtrl = new SimController(simCanvas, this.sim1);
        this.advance1 = new SimpleAdvance(this.sim1);
        this.advance2 = new SimpleAdvance(this.sim2);
        this.simRect = new DoubleRect(-2, -2.2, 2, 1.5);
        this.simView = new SimView('simView', this.simRect);
        this.displayList = this.simView.getDisplayList();
        simCanvas.addView(this.simView);
        this.statusView = new SimView('status', new DoubleRect(-10, -10, 10, 10));
        simCanvas.addView(this.statusView);
        this.axes = CommonControls.makeAxes(this.simView, true);
        this.simRun = new SimRunner(this.advance1);
        this.simRun.addStrategy(this.advance2);
        this.simRun.addCanvas(simCanvas);
        this.clock = this.simRun.getClock();
        const displayRod2 = new DisplayLine(this.simList2.getConcreteLine('rod'));
        this.displayList.add(displayRod2);
        this.bob2 = this.simList2.getPointMass('bob');
        const displayBob2 = new DisplayShape(this.bob2);
        displayBob2.setFillStyle('red');
        this.displayList.add(displayBob2);
        displayBob2.setDragable(false);
        this.bob1 = this.simList.getPointMass('bob');
        const displayBob = new DisplayShape(this.bob1);
        displayBob.setFillStyle('blue');
        this.displayList.add(displayBob);
        const displayRod = new DisplayLine(this.simList.getConcreteLine('rod'));
        this.displayList.add(displayRod);
        const displayDrive = new DisplayArc(this.simList.getArc('drive'));
        this.displayList.add(displayDrive);
        this.addControl(CommonControls.makePlaybackControls(this.simRun));
        let pn = new ParameterNumber(this, ComparePendulumApp.en.ANGLE_DELTA, ComparePendulumApp.i18n.ANGLE_DELTA, () => this.getAngleDelta(), a => this.setAngleDelta(a))
            .setDecimalPlaces(7);
        this.addParameter(pn);
        this.addControl(new NumericControl(pn));
        pn = this.sim1.getParameterNumber(PendulumSim.en.LENGTH);
        this.addControl(new SliderControl(pn, 0.1, 10.1, true));
        pn = this.sim1.getParameterNumber(PendulumSim.en.DAMPING);
        this.addControl(new SliderControl(pn, 0, 10, false));
        pn = this.sim1.getParameterNumber(PendulumSim.en.MASS);
        this.addControl(new SliderControl(pn, 0.2, 10.2, true));
        pn = this.sim1.getParameterNumber(PendulumSim.en.DRIVE_AMPLITUDE);
        this.addControl(new SliderControl(pn, 0, 10, false));
        pn = this.sim1.getParameterNumber(PendulumSim.en.DRIVE_FREQUENCY);
        this.addControl(new SliderControl(pn, 0, 10, false));
        pn = this.sim1.getParameterNumber(PendulumSim.en.GRAVITY);
        this.addControl(new SliderControl(pn, 0, 20, false));
        this.energyGraph = new EnergyBarGraph(this.sim1);
        this.showEnergyParam = CommonControls.makeShowEnergyParam(this.energyGraph, this.statusView, this);
        this.addControl(new CheckBoxControl(this.showEnergyParam));
        this.displayClock = new DisplayClock(() => this.sim1.getTime(), () => this.clock.getRealTime(), 2, 2);
        this.displayClock.setPosition(new Vector(8, 4));
        let pb = CommonControls.makeShowClockParam(this.displayClock, this.statusView, this);
        this.addControl(new CheckBoxControl(pb));
        const panzoom_simview = CommonControls.makePanZoomControls(this.simView, true, () => this.simView.setSimRect(this.simRect));
        this.layout.getSimDiv().appendChild(panzoom_simview);
        pb = CommonControls.makeShowPanZoomParam(panzoom_simview, this);
        pb.setValue(false);
        this.addControl(new CheckBoxControl(pb));
        pn = this.simRun.getParameterNumber(SimRunner.en.TIME_STEP);
        this.addControl(new NumericControl(pn));
        pn = this.simRun.getClock().getParameterNumber(Clock.en.TIME_RATE);
        this.addControl(new NumericControl(pn));
        const bm = CommonControls.makeBackgroundMenu(this.layout.getSimCanvas());
        this.addControl(bm);
        const line1 = new GraphLine('GRAPH_LINE_1', va1);
        line1.setXVariable(0);
        line1.setYVariable(1);
        line1.setColor('blue');
        line1.setDrawingMode("dots");
        const line2 = new GraphLine('GRAPH_LINE_2', va2);
        line2.setXVariable(0);
        line2.setYVariable(1);
        line2.setColor('red');
        line2.setDrawingMode("dots");
        const paramY = line1.getParameterNumber(GraphLine.en.Y_VARIABLE);
        const paramX = line1.getParameterNumber(GraphLine.en.X_VARIABLE);
        const py2 = line2.getParameterNumber(GraphLine.en.Y_VARIABLE);
        const px2 = line2.getParameterNumber(GraphLine.en.X_VARIABLE);
        new GenericObserver(line1, evt => {
            if (evt == paramY) {
                py2.setValue(paramY.getValue());
            }
            else if (evt == paramX) {
                px2.setValue(paramX.getValue());
            }
        }, 'keep line2\'s X and Y variable in sync with line1');
        this.graph = new CompareGraph(line1, line2, this.layout.getGraphCanvas(), this.layout.getGraphControls(), this.layout.getGraphDiv(), this.simRun);
        const timeLine1 = new GraphLine('TIME_LINE_1', va1);
        timeLine1.setYVariable(0);
        timeLine1.setColor('blue');
        timeLine1.setDrawingMode("dots");
        const timeLine2 = new GraphLine('TIME_LINE_2', va2);
        timeLine2.setYVariable(0);
        timeLine2.setColor('red');
        timeLine2.setDrawingMode("dots");
        const timeParamY = timeLine1.getParameterNumber(GraphLine.en.Y_VARIABLE);
        const time_py2 = timeLine2.getParameterNumber(GraphLine.en.Y_VARIABLE);
        new GenericObserver(timeLine1, evt => {
            if (evt == timeParamY) {
                time_py2.setValue(timeParamY.getValue());
            }
        }, 'keep timeLine2\'s Y variable in sync with timeLine1');
        this.timeGraph = new CompareTimeGraph(timeLine1, timeLine2, this.layout.getTimeGraphCanvas(), this.layout.getTimeGraphControls(), this.layout.getTimeGraphDiv(), this.simRun);
        new GenericObserver(this.sim1, (evt) => {
            if (evt instanceof ParameterNumber) {
                const pn = this.sim2.getParameterNumber(evt.getName());
                pn.setValue(evt.getValue());
            }
            else if (evt.nameEquals('START_DRAG')) {
                this.sim2.setIsDragging(true);
            }
            else if (evt.nameEquals('MOUSE_DRAG')) {
                va2.setValue(0, va1.getValue(0) + this.angleDelta);
                va2.setValue(1, va1.getValue(1));
                this.sim2.modifyObjects();
            }
            else if (evt.nameEquals('FINISH_DRAG')) {
                this.sim2.setIsDragging(false);
                va2.setTime(0);
                this.sim2.saveInitialState();
                va1.setTime(0);
                this.sim1.saveInitialState();
                this.simRun.reset();
            }
            else if (evt.nameEquals('RESET')) {
                this.sim2.reset();
            }
        }, 'sim2 follows sim1 when mouse dragging');
        let subjects = [
            this,
            this.sim1,
            this.sim2,
            this.simRun,
            this.clock,
            this.simView,
            this.statusView,
            this.sim1.getVarsList(),
            this.sim2.getVarsList()
        ];
        subjects = subjects.concat(this.layout.getSubjects(), this.graph.getSubjects(), this.timeGraph.getSubjects());
        this.easyScript = CommonControls.makeEasyScript(subjects, [], this.simRun, this.terminal);
        this.addControl(CommonControls.makeURLScriptButton(this.easyScript, this.simRun));
        this.graph.addControl(CommonControls.makeURLScriptButton(this.easyScript, this.simRun));
        this.timeGraph.addControl(CommonControls.makeURLScriptButton(this.easyScript, this.simRun));
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', sim1: ' + this.sim1.toStringShort()
            + ', sim2: ' + this.sim2.toStringShort()
            + ', terminal: ' + this.terminal
            + ', graph: ' + this.graph
            + ', timeGraph: ' + this.timeGraph
            + super.toString();
    }
    ;
    getClassName() {
        return 'ComparePendulumApp';
    }
    ;
    defineNames(myName) {
        this.terminal.addRegex('advance1|advance2|axes|clock|displayClock|displayList'
            + '|energyGraph|graph|layout|sim1|sim2|simCtrl|simList|simList2'
            + '|simRect|simRun|simView|statusView|timeGraph|easyScript|terminal', myName + '.');
        this.terminal.addRegex('simCanvas', myName + '.layout.');
    }
    ;
    addControl(control) {
        this.layout.addControl(control);
    }
    ;
    setup() {
        this.clock.resume();
        this.terminal.parseURL();
        this.sim1.saveInitialState();
        this.sim1.modifyObjects();
    }
    ;
    start() {
        this.simRun.startFiring();
    }
    ;
    eval(script, output = true) {
        return this.terminal.eval(script, output);
    }
    ;
    getAngleDelta() {
        return this.angleDelta;
    }
    ;
    setAngleDelta(value) {
        if (this.angleDelta != value) {
            this.angleDelta = value;
            this.simRun.reset();
            const startAngle = this.sim1.getVarsList().getValue(0);
            this.sim2.getVarsList().setValue(0, startAngle + this.angleDelta);
            this.sim2.saveInitialState();
            this.broadcastParameter(ComparePendulumApp.en.ANGLE_DELTA);
        }
    }
    ;
}
ComparePendulumApp.en = {
    ANGLE_DELTA: 'angle difference'
};
ComparePendulumApp.de_strings = {
    ANGLE_DELTA: 'Winkeldifferenz'
};
ComparePendulumApp.i18n = Util.LOCALE === 'de' ? ComparePendulumApp.de_strings : ComparePendulumApp.en;
Util.defineGlobal('sims$pendulum$ComparePendulumApp', ComparePendulumApp);
