import { AbstractODESim } from '../../lab/model/ODESim.js';
import { Arc } from '../../lab/model/Arc.js';
import { ConcreteLine } from '../../lab/model/ConcreteLine.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { GenericEvent, ParameterBoolean, ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TH = 0;
const TH_P = 1;
const TIME = 2;
const TH_PP = 3;
const KE = 4;
const PE = 5;
const TE = 6;
export class PendulumSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.pivot_ = Vector.ORIGIN;
        this.length_ = 1;
        this.gravity_ = 1;
        this.damping_ = 0.5;
        this.frequency_ = 2 / 3;
        this.amplitude_ = 1.15;
        this.limitAngle_ = true;
        this.rod_ = new ConcreteLine('rod');
        this.bob_ = PointMass.makeCircle(0.2, 'bob');
        this.potentialOffset_ = 0;
        this.drive_ = new Arc('drive', -Math.PI / 2, 0.5 * this.amplitude_, this.pivot_);
        this.isDragging_ = false;
        const var_names = [
            PendulumSim.en.ANGLE,
            PendulumSim.en.ANGULAR_VELOCITY,
            VarsList.en.TIME,
            PendulumSim.en.ANGULAR_ACCEL,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY
        ];
        const i18n_names = [
            PendulumSim.i18n.ANGLE,
            PendulumSim.i18n.ANGULAR_VELOCITY,
            VarsList.i18n.TIME,
            PendulumSim.i18n.ANGULAR_ACCEL,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY
        ];
        this.bob_.setMass(1.0);
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(TH_PP, KE, PE, TE);
        this.getSimList().add(this.rod_, this.drive_, this.bob_);
        this.modifyObjects();
        this.saveInitialState();
        this.addParameter(new ParameterNumber(this, PendulumSim.en.LENGTH, PendulumSim.i18n.LENGTH, () => this.getLength(), a => this.setLength(a)));
        this.addParameter(new ParameterNumber(this, PendulumSim.en.DAMPING, PendulumSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, PendulumSim.en.MASS, PendulumSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, PendulumSim.en.DRIVE_AMPLITUDE, PendulumSim.i18n.DRIVE_AMPLITUDE, () => this.getDriveAmplitude(), a => this.setDriveAmplitude(a)));
        this.addParameter(new ParameterNumber(this, PendulumSim.en.DRIVE_FREQUENCY, PendulumSim.i18n.DRIVE_FREQUENCY, () => this.getDriveFrequency(), a => this.setDriveFrequency(a)));
        this.addParameter(new ParameterNumber(this, PendulumSim.en.GRAVITY, PendulumSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterBoolean(this, PendulumSim.en.LIMIT_ANGLE, PendulumSim.i18n.LIMIT_ANGLE, () => this.getLimitAngle(), a => this.setLimitAngle(a)));
        this.addParameter(new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a))
            .setLowerLimit(Number.NEGATIVE_INFINITY)
            .setSignifDigits(5));
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', length_: ' + Util.NF(this.length_)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', damping_: ' + Util.NF(this.damping_)
            + ', frequency_: ' + Util.NF(this.frequency_)
            + ', amplitude_: ' + Util.NF(this.amplitude_)
            + ', limitAngle_: ' + this.limitAngle_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + ', pivot_: ' + this.pivot_
            + ', rod_: ' + this.rod_
            + ', bob_: ' + this.bob_
            + super.toString();
    }
    ;
    getClassName() {
        return 'PendulumSim';
    }
    ;
    getEnergyInfo() {
        const ke = this.bob_.getKineticEnergy();
        const y = this.bob_.getPosition().getY();
        const pe = this.gravity_ * this.bob_.getMass() * (y + this.length_);
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        if (this.limitAngle_) {
            const angle = Util.limitAngle(vars[TH]);
            if (angle != vars[TH]) {
                this.getVarsList().setValue(0, angle, false);
                vars[TH] = angle;
            }
        }
        this.moveObjects(vars);
        const rate = new Array(vars.length);
        this.evaluate(vars, rate, 0);
        vars[TH_PP] = rate[TH_P];
        const ei = this.getEnergyInfo();
        vars[KE] = ei.getTranslational();
        vars[PE] = ei.getPotential();
        vars[TE] = ei.getTotalEnergy();
        va.setValues(vars, true);
    }
    ;
    moveObjects(vars) {
        const angle = vars[TH];
        const sinAngle = Math.sin(angle);
        const cosAngle = Math.cos(angle);
        const len = this.length_;
        this.bob_.setPosition(new Vector(this.pivot_.getX() + len * sinAngle, this.pivot_.getY() - len * cosAngle));
        const vx = vars[TH_P] * len * cosAngle;
        const vy = vars[TH_P] * len * sinAngle;
        this.bob_.setVelocity(new Vector(vx, vy));
        this.rod_.setStartPoint(this.pivot_);
        this.rod_.setEndPoint(this.bob_.getPosition());
        let t = this.frequency_ * vars[TIME];
        t = 180 * t / Math.PI;
        t = t - 360 * Math.floor(t / 360);
        if ((t >= 0) && (t <= 180)) {
            t = 90 - t;
        }
        else {
            t = t - 270;
        }
        this.drive_.setAngle(Math.PI * t / 180.0);
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.bob_) {
            this.isDragging_ = true;
            this.broadcast(new GenericEvent(this, 'START_DRAG', simObject));
            return true;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const vars = va.getValues();
        if (simObject == this.bob_) {
            const p = location.subtract(offset).subtract(this.pivot_);
            vars[TH] = Math.PI / 2 + Math.atan2(p.getY(), p.getX());
            vars[TH_P] = 0;
            va.setValues(vars);
            this.moveObjects(vars);
            this.broadcast(new GenericEvent(this, 'MOUSE_DRAG', simObject));
        }
    }
    ;
    finishDrag(simObject, _location, _offset) {
        if (this.isDragging_) {
            this.isDragging_ = false;
            this.broadcast(new GenericEvent(this, 'FINISH_DRAG', simObject));
        }
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        if (!this.isDragging_) {
            change[TH] = vars[TH_P];
            const len = this.length_;
            let dd = -(this.gravity_ / len) * Math.sin(vars[TH]);
            const mlsq = this.bob_.getMass() * len * len;
            dd += -(this.damping_ / mlsq) * vars[TH_P];
            dd += (this.amplitude_ / mlsq) * Math.cos(this.frequency_ * vars[TIME]);
            change[TH_P] = dd;
        }
        return null;
    }
    ;
    getIsDragging() {
        return this.isDragging_;
    }
    ;
    setIsDragging(value) {
        this.isDragging_ = value;
    }
    ;
    getMass() {
        return this.bob_.getMass();
    }
    ;
    setMass(value) {
        this.bob_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(PendulumSim.en.MASS);
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(PendulumSim.en.GRAVITY);
    }
    ;
    getDriveFrequency() {
        return this.frequency_;
    }
    ;
    setDriveFrequency(value) {
        this.frequency_ = value;
        this.broadcastParameter(PendulumSim.en.DRIVE_FREQUENCY);
    }
    ;
    getDriveAmplitude() {
        return this.amplitude_;
    }
    ;
    setDriveAmplitude(value) {
        this.amplitude_ = value;
        this.drive_.setRadius(Math.min(2 * this.length_, 0.5 * this.amplitude_));
        this.modifyObjects();
        this.broadcastParameter(PendulumSim.en.DRIVE_AMPLITUDE);
    }
    ;
    getLength() {
        return this.length_;
    }
    ;
    setLength(value) {
        this.length_ = value;
        this.drive_.setRadius(Math.min(2 * this.length_, 0.5 * this.amplitude_));
        this.getVarsList().incrSequence(KE, PE, TE);
        this.modifyObjects();
        this.broadcastParameter(PendulumSim.en.LENGTH);
    }
    ;
    getLimitAngle() {
        return this.limitAngle_;
    }
    ;
    setLimitAngle(value) {
        this.limitAngle_ = value;
        this.broadcastParameter(PendulumSim.en.LIMIT_ANGLE);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(PendulumSim.en.DAMPING);
    }
    ;
    getPivot() {
        return this.pivot_;
    }
    ;
    setPivot(value) {
        this.pivot_ = value;
        this.drive_.setCenter(value);
        this.modifyObjects();
    }
    ;
}
PendulumSim.en = {
    DRIVE_AMPLITUDE: 'drive amplitude',
    ANGLE: 'angle',
    ANGULAR_ACCEL: 'angle acceleration',
    ANGULAR_VELOCITY: 'angle velocity',
    DAMPING: 'damping',
    DRIVE_FREQUENCY: 'drive frequency',
    GRAVITY: 'gravity',
    LENGTH: 'length',
    LIMIT_ANGLE: 'limit angle',
    MASS: 'mass',
    TIME: 'time'
};
PendulumSim.de_strings = {
    DRIVE_AMPLITUDE: 'Antriebsamplitude',
    ANGLE: 'Winkel',
    ANGULAR_ACCEL: 'Winkelbeschleunigung',
    ANGULAR_VELOCITY: 'Winkelgeschwindigkeit',
    DAMPING: 'Dämpfung',
    DRIVE_FREQUENCY: 'Antriebsfrequenz',
    GRAVITY: 'Gravitation',
    LENGTH: 'Länge',
    LIMIT_ANGLE: 'Winkel begrenzen',
    MASS: 'Masse',
    TIME: 'Zeit'
};
PendulumSim.i18n = Util.LOCALE === 'de' ? PendulumSim.de_strings : PendulumSim.en;
Util.defineGlobal('sims$pendulum$PendulumSim', PendulumSim);
