import { AbstractODESim } from '../../lab/model/ODESim.js';
import { ConcreteLine } from '../../lab/model/ConcreteLine.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { UtilEngine } from '../../lab/engine2D/UtilEngine.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const X = 0;
const XP = 1;
const Y = 2;
const YP = 3;
const ANGLE = 4;
const ANGLEP = 5;
const TIME = 6;
const KE = 7;
const PE = 8;
const TE = 9;
export class ReactionPendulumSim extends AbstractODESim {
    constructor(length, radius, startAngle, opt_name, opt_simList) {
        super(opt_name, opt_simList);
        this.mass_ = 1;
        this.gravity_ = 1;
        this.damping_ = 0;
        this.potentialOffset_ = 0;
        this.rod_ = new ConcreteLine('rod');
        const var_names = [
            ReactionPendulumSim.en.X_POSITION,
            ReactionPendulumSim.en.X_VELOCITY,
            ReactionPendulumSim.en.Y_POSITION,
            ReactionPendulumSim.en.Y_VELOCITY,
            ReactionPendulumSim.en.ANGLE,
            ReactionPendulumSim.en.ANGULAR_VELOCITY,
            VarsList.en.TIME,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY
        ];
        const i18n_names = [
            ReactionPendulumSim.i18n.X_POSITION,
            ReactionPendulumSim.i18n.X_VELOCITY,
            ReactionPendulumSim.i18n.Y_POSITION,
            ReactionPendulumSim.i18n.Y_VELOCITY,
            ReactionPendulumSim.i18n.ANGLE,
            ReactionPendulumSim.i18n.ANGULAR_VELOCITY,
            VarsList.i18n.TIME,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.radius_ = radius;
        this.length_ = length;
        this.getSimList().add(this.rod_);
        this.config(length, radius, startAngle);
        this.addParameter(new ParameterNumber(this, ReactionPendulumSim.en.DAMPING, ReactionPendulumSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, ReactionPendulumSim.en.MASS, ReactionPendulumSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, ReactionPendulumSim.en.GRAVITY, ReactionPendulumSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        let pn;
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', damping_: ' + Util.NF(this.damping_)
            + ', length_: ' + Util.NF(this.length_)
            + ', rod_: ' + this.rod_
            + ', bob_: ' + this.bob_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'ReactionPendulumSim';
    }
    ;
    config(length, radius, startAngle) {
        this.length_ = length;
        this.radius_ = radius;
        const va = this.getVarsList();
        const vars = va.getValues();
        vars[ANGLE] = startAngle;
        vars[X] = length * Math.sin(vars[ANGLE]);
        vars[Y] = -length * Math.cos(vars[ANGLE]);
        vars[XP] = vars[YP] = vars[ANGLEP] = 0;
        vars[TIME] = 0;
        va.setValues(vars);
        if (this.bob_ !== undefined) {
            this.getSimList().remove(this.bob_);
        }
        this.bob_ = Shapes.makeBall(radius);
        this.bob_.setMass(this.mass_);
        this.getSimList().add(this.bob_);
        this.saveInitialState();
    }
    ;
    getEnergyInfo() {
        const vars = this.getVarsList().getValues();
        const ke = 0.5 * this.mass_ * (vars[XP] * vars[XP] + vars[YP] * vars[YP]);
        Util.assert(!Util.veryDifferent(ke, this.bob_.translationalEnergy()));
        const I = this.mass_ * this.radius_ * this.radius_ / 2;
        Util.assert(!Util.veryDifferent(I, this.bob_.momentAboutCM()));
        const re = 0.5 * I * vars[ANGLEP] * vars[ANGLEP];
        Util.assert(!Util.veryDifferent(re, this.bob_.rotationalEnergy()));
        const pe = this.gravity_ * this.mass_ * (vars[Y] + this.length_);
        return new EnergyInfo(pe + this.potentialOffset_, ke, re);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        this.bob_.setPosition(new Vector(vars[X], vars[Y]), vars[ANGLE]);
        this.bob_.setVelocity(new Vector(vars[XP], vars[YP]), vars[ANGLEP]);
        this.rod_.setStartPoint(Vector.ORIGIN);
        this.rod_.setEndPoint(this.bob_.getPosition());
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        const m = this.mass_;
        change[X] = vars[XP];
        change[XP] = -this.damping_ * vars[XP];
        change[Y] = vars[YP];
        change[YP] = -this.gravity_ - this.damping_ * vars[YP];
        change[ANGLE] = vars[ANGLEP];
        change[ANGLEP] = 0;
        const len = this.length_;
        const I = m * (this.radius_ * this.radius_ / 2.0);
        const n0x = 0;
        const n0y = -1;
        const n1x = -1;
        const n1y = 0;
        const rx = -len * Math.sin(vars[ANGLE]);
        const ry = len * Math.cos(vars[ANGLE]);
        const vx = vars[XP];
        const vy = vars[YP];
        const w = vars[ANGLEP];
        const A = [new Float64Array(3), new Float64Array(3)];
        const B = [0, 0];
        const f = [0, 0];
        let nx, ny, nix, niy, b;
        nx = n0x;
        ny = n0y;
        b = nx * (change[XP] - change[ANGLEP] * ry - w * w * rx) + ny * (change[YP] + change[ANGLEP] * rx - w * w * ry);
        B[0] = b;
        nx = n1x;
        ny = n1y;
        b = nx * (change[XP] - change[ANGLEP] * ry - w * w * rx) + ny * (change[YP] + change[ANGLEP] * rx - w * w * ry);
        B[1] = b;
        nx = n0x;
        ny = n0y;
        nix = n0x;
        niy = n0y;
        A[0][0] = nix * (nx / m - ry * (rx * ny - ry * nx) / I) + niy * (ny / m + rx * (rx * ny - ry * nx) / I);
        nx = n1x;
        ny = n1y;
        nix = n0x;
        niy = n0y;
        A[0][1] = nix * (nx / m - ry * (rx * ny - ry * nx) / I) + niy * (ny / m + rx * (rx * ny - ry * nx) / I);
        nx = n0x;
        ny = n0y;
        nix = n1x;
        niy = n1y;
        A[1][0] = nix * (nx / m - ry * (rx * ny - ry * nx) / I) + niy * (ny / m + rx * (rx * ny - ry * nx) / I);
        nx = n1x;
        ny = n1y;
        nix = n1x;
        niy = n1y;
        A[1][1] = nix * (nx / m - ry * (rx * ny - ry * nx) / I) + niy * (ny / m + rx * (rx * ny - ry * nx) / I);
        B[0] = -B[0];
        B[1] = -B[1];
        const err = UtilEngine.matrixSolve4(A, f, B);
        if (err != -1) {
            throw err;
        }
        nx = n0x;
        ny = n0y;
        let Fx, Fy;
        Fx = f[0] * nx;
        Fy = f[0] * ny;
        change[XP] += f[0] * nx / m;
        change[YP] += f[0] * ny / m;
        change[ANGLEP] += (rx * f[0] * ny - ry * f[0] * nx) / I;
        nx = n1x;
        ny = n1y;
        Fx += f[1] * nx;
        Fy += f[1] * ny;
        change[XP] += f[1] * nx / m;
        change[YP] += f[1] * ny / m;
        change[ANGLEP] += (rx * f[1] * ny - ry * f[1] * nx) / I;
        this.showForce(Fx, 0);
        this.showForce(0, Fy);
        return null;
    }
    ;
    showForce(fx, fy) {
        const v = new ConcreteLine('contact_force', Vector.ORIGIN, new Vector(fx, fy));
        v.setExpireTime(this.getTime());
        this.getSimList().add(v);
    }
    ;
    getMass() {
        return this.mass_;
    }
    ;
    setMass(value) {
        if (this.mass_ != value) {
            this.mass_ = value;
            this.bob_.setMass(value);
            this.getVarsList().incrSequence(KE, PE, TE);
            this.broadcastParameter(ReactionPendulumSim.en.MASS);
        }
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(ReactionPendulumSim.en.GRAVITY);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(ReactionPendulumSim.en.DAMPING);
    }
    ;
}
ReactionPendulumSim.en = {
    X_POSITION: 'X position',
    Y_POSITION: 'Y position',
    X_VELOCITY: 'X velocity',
    Y_VELOCITY: 'Y velocity',
    ANGLE: 'angle',
    ANGULAR_VELOCITY: 'angle velocity',
    START_ANGLE: 'start angle',
    DAMPING: 'damping',
    GRAVITY: 'gravity',
    LENGTH: 'length',
    MASS: 'mass',
    TIME: 'time',
    RADIUS: 'radius'
};
ReactionPendulumSim.de_strings = {
    X_POSITION: 'X Position',
    Y_POSITION: 'Y Position',
    X_VELOCITY: 'X Geschwindigkeit',
    Y_VELOCITY: 'Y Geschwindigkeit',
    ANGLE: 'Winkel',
    ANGULAR_VELOCITY: 'Winkelgeschwindigkeit',
    START_ANGLE: 'anfangs Winkel',
    DAMPING: 'Dämpfung',
    GRAVITY: 'Gravitation',
    LENGTH: 'Länge',
    MASS: 'Masse',
    TIME: 'Zeit',
    RADIUS: 'Radius'
};
ReactionPendulumSim.i18n = Util.LOCALE === 'de' ? ReactionPendulumSim.de_strings : ReactionPendulumSim.en;
Util.defineGlobal('sims$pendulum$ReactionPendulumSim', ReactionPendulumSim);
