import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { Joint } from '../../lab/engine2D/Joint.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Scrim } from '../../lab/engine2D/Scrim.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const THETA1 = 0;
const THETA1P = 1;
const THETA2 = 2;
const THETA2P = 3;
const KE = 4;
const PE = 5;
const TE = 6;
const TIME = 7;
export class RigidDoublePendulumSim extends AbstractODESim {
    constructor(parts, opt_name, opt_simList) {
        super(opt_name, opt_simList);
        this.potentialOffset_ = 0;
        this.gravity_ = 9.8;
        const var_names = [
            RigidDoublePendulumSim.en.ANGLE_1,
            RigidDoublePendulumSim.en.ANGLE_1_VELOCITY,
            RigidDoublePendulumSim.en.ANGLE_2,
            RigidDoublePendulumSim.en.ANGLE_2_VELOCITY,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME
        ];
        const i18n_names = [
            RigidDoublePendulumSim.i18n.ANGLE_1,
            RigidDoublePendulumSim.i18n.ANGLE_1_VELOCITY,
            RigidDoublePendulumSim.i18n.ANGLE_2,
            RigidDoublePendulumSim.i18n.ANGLE_2_VELOCITY,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.pendulum1_ = parts.bodies[0];
        this.pendulum2_ = parts.bodies[1];
        this.pivot1_ = parts.joints[0];
        this.pivot2_ = parts.joints[1];
        if (!(this.pivot1_.getBody1() instanceof Scrim)
            || this.pivot1_.getBody2() != this.pendulum1_) {
            throw 'parts error 1';
        }
        if (this.pivot2_.getBody1() != this.pendulum1_
            || this.pivot2_.getBody2() != this.pendulum2_) {
            throw 'parts error 2';
        }
        if (!this.pivot1_.getAttach1().equals(Vector.ORIGIN)) {
            throw 'parts error 3';
        }
        this.gamma1_ = RigidDoublePendulumSim.getGamma(this.pendulum1_, this.pivot1_);
        this.gamma2_ = RigidDoublePendulumSim.getGamma(this.pendulum2_, this.pivot2_);
        this.omega1_ = this.pendulum1_.getAngle();
        this.omega2_ = this.pendulum2_.getAngle();
        const theta1 = this.omega1_ + this.gamma1_;
        const theta1_velocity = this.pendulum1_.getAngularVelocity();
        const theta2 = this.omega2_ + this.gamma2_;
        const theta2_velocity = this.pendulum2_.getAngularVelocity();
        const c1 = this.pendulum1_.getCenterOfMass();
        const attach1_0 = this.pivot1_.getAttach2();
        const r1 = c1.subtract(attach1_0);
        const l1 = this.pivot2_.getAttach1().subtract(attach1_0);
        const c2 = this.pendulum2_.getCenterOfMass();
        const r2 = c2.subtract(this.pivot2_.getAttach2());
        this.R1_ = r1.length();
        this.L1_ = l1.length();
        this.R2_ = r2.length();
        this.phi_ = r1.angleTo(l1);
        const vars = this.getVarsList().getValues();
        vars[THETA1] = 0;
        vars[THETA1P] = 0;
        vars[THETA2] = 0;
        vars[THETA2P] = 0;
        this.getVarsList().setValues(vars);
        this.modifyObjects();
        this.potentialOffset_ = -this.getEnergyInfo().getPotential();
        vars[THETA1] = theta1;
        vars[THETA1P] = theta1_velocity;
        vars[THETA2] = theta2;
        vars[THETA2P] = theta2_velocity;
        this.getVarsList().setValues(vars);
        this.saveInitialState();
        let pn;
        this.addParameter(new ParameterNumber(this, RigidDoublePendulumSim.en.GRAVITY, RigidDoublePendulumSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(pn = new ParameterNumber(this, RigidDoublePendulumSim.en.ANGLE_1, RigidDoublePendulumSim.i18n.ANGLE_1, () => this.getAngle1(), a => this.setAngle1(a)));
        pn.setLowerLimit(-Math.PI);
        pn.setUpperLimit(Math.PI);
        this.addParameter(pn = new ParameterNumber(this, RigidDoublePendulumSim.en.ANGLE_2, RigidDoublePendulumSim.i18n.ANGLE_2, () => this.getAngle2(), a => this.setAngle2(a)));
        pn.setLowerLimit(-Math.PI);
        pn.setUpperLimit(Math.PI);
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
        this.getSimList().add(this.pendulum1_, this.pendulum2_, this.pivot1_, this.pivot2_);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', gamma1_: ' + Util.NF(this.gamma1_)
            + ', gamma2_: ' + Util.NF(this.gamma2_)
            + ', pendulum1_: ' + this.pendulum1_
            + ', pendulum2_: ' + this.pendulum2_
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'RigidDoublePendulumSim';
    }
    ;
    static makeOffset(theta1, theta2, pivot) {
        pivot = pivot || Vector.ORIGIN;
        const p1 = Shapes.makeBlock2(0.3, 1.0, RigidDoublePendulumSim.en.PENDULUM + 1, RigidDoublePendulumSim.i18n.PENDULUM + 1);
        p1.setCenterOfMass(new Vector(p1.getWidth() / 3.0, p1.getHeight() * 0.3));
        p1.setPosition(new Vector(0, 0), theta1);
        const p2 = Shapes.makeBlock2(0.3, 1.0, RigidDoublePendulumSim.en.PENDULUM + 2, RigidDoublePendulumSim.i18n.PENDULUM + 2);
        p2.setPosition(new Vector(0, 0), theta2);
        const scrim = Scrim.getScrim();
        const j1 = new Joint(scrim, pivot, p1, new Vector(0.5 * p1.getWidth(), 0.85 * p1.getHeight()), 0, Vector.NORTH);
        const j2 = new Joint(p1, new Vector(0.85 * p1.getWidth(), 0.15 * p1.getHeight()), p2, new Vector(0.15 * p2.getWidth(), 0.85 * p2.getHeight()), 0, Vector.NORTH);
        const j1_1 = new Joint(scrim, pivot, p1, j1.getAttach2(), 0, Vector.EAST);
        const j2_1 = new Joint(p1, j2.getAttach1(), p2, j2.getAttach2(), 0, Vector.EAST);
        return { bodies: [p1, p2], joints: [j1, j2, j1_1, j2_1] };
    }
    ;
    static makeCentered(theta1, theta2, pivot) {
        pivot = pivot || Vector.ORIGIN;
        const p1 = Shapes.makeBlock2(0.3, 1.0, RigidDoublePendulumSim.en.PENDULUM + 1, RigidDoublePendulumSim.i18n.PENDULUM + 1);
        p1.setPosition(new Vector(0, 0), theta1);
        const p2 = Shapes.makeBlock2(0.3, 1.0, RigidDoublePendulumSim.en.PENDULUM + 2, RigidDoublePendulumSim.i18n.PENDULUM + 2);
        p2.setPosition(new Vector(0, 0), theta2);
        const scrim = Scrim.getScrim();
        const j1 = new Joint(scrim, pivot, p1, new Vector(0.5 * p1.getWidth(), 0.85 * p1.getHeight()), 0, Vector.NORTH);
        const j2 = new Joint(p1, new Vector(0.5 * p1.getWidth(), 0.15 * p1.getHeight()), p2, new Vector(0.5 * p2.getWidth(), 0.85 * p2.getHeight()), 0, Vector.NORTH);
        const j1_1 = new Joint(scrim, pivot, p1, j1.getAttach2(), 0, Vector.EAST);
        const j2_1 = new Joint(p1, j2.getAttach1(), p2, j2.getAttach2(), 0, Vector.EAST);
        return { bodies: [p1, p2], joints: [j1, j2, j1_1, j2_1] };
    }
    ;
    static getGamma(pendulum, pivot) {
        let attach;
        if (pivot.getBody1() == pendulum) {
            attach = pivot.getAttach1();
        }
        else if (pivot.getBody2() == pendulum) {
            attach = pivot.getAttach2();
        }
        else {
            throw 'parts error 4';
        }
        const v = attach.subtract(pendulum.getCenterOfMass());
        return v.getAngle() - Math.PI / 2;
    }
    ;
    getEnergyInfo() {
        const p1 = this.pendulum1_;
        const p2 = this.pendulum2_;
        const ke = p1.translationalEnergy() + p2.translationalEnergy();
        const re = p1.rotationalEnergy() + p2.rotationalEnergy();
        const pe = this.gravity_ * p1.getMass() * p1.getPosition().getY()
            + this.gravity_ * p2.getMass() * p2.getPosition().getY();
        return new EnergyInfo(pe + this.potentialOffset_, ke, re);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational() + ei.getRotational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        const sin_th1 = Math.sin(vars[THETA1]);
        const cos_th1 = Math.cos(vars[THETA1]);
        const sin_th2 = Math.sin(vars[THETA2]);
        const cos_th2 = Math.cos(vars[THETA2]);
        const sin_ph1 = Math.sin(vars[THETA1] + this.phi_);
        const cos_ph1 = Math.cos(vars[THETA1] + this.phi_);
        const R1 = this.R1_;
        const R2 = this.R2_;
        const L1 = this.L1_;
        this.pendulum1_.setPosition(new Vector(R1 * sin_th1, -R1 * cos_th1), vars[THETA1] - this.gamma1_);
        this.pendulum1_.setVelocity(new Vector(vars[THETA1P] * R1 * cos_th1, vars[THETA1P] * R1 * sin_th1), vars[THETA1P]);
        this.pendulum2_.setPosition(new Vector(L1 * sin_ph1 + R2 * sin_th2, -L1 * cos_ph1 - R2 * cos_th2), vars[THETA2] - this.gamma2_);
        this.pendulum2_.setVelocity(new Vector(vars[THETA1P] * L1 * cos_ph1 + vars[THETA2P] * R2 * cos_th2, vars[THETA1P] * L1 * sin_ph1 + vars[THETA2P] * R2 * sin_th2), vars[THETA2P]);
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        const th1 = vars[THETA1];
        const dth1 = vars[THETA1P];
        const th2 = vars[THETA2];
        const dth2 = vars[THETA2P];
        const m1 = this.pendulum1_.getMass();
        const m2 = this.pendulum2_.getMass();
        const I1 = this.pendulum1_.momentAboutCM();
        const I2 = this.pendulum2_.momentAboutCM();
        const R1 = this.R1_;
        const R2 = this.R2_;
        const L1 = this.L1_;
        const g = this.gravity_;
        const phi = this.phi_;
        change[THETA1] = dth1;
        change[THETA1P] = -((2 * g * m1 * R1 * (I2 + m2 * R2 * R2) * Math.sin(th1) +
            L1 * m2 * (g * (2 * I2 + m2 * R2 * R2) * Math.sin(th1 + phi) +
                R2 * (g * m2 * R2 * Math.sin(th1 - 2 * th2 + phi) +
                    2 * (dth2 * dth2 * (I2 + m2 * R2 * R2) +
                        dth1 * dth1 * L1 * m2 * R2 * Math.cos(th1 - th2 + phi)) * Math.sin(th1 - th2 + phi)))) /
            (2 * I2 * L1 * L1 * m2 + 2 * I2 * m1 * R1 * R1 + L1 * L1 * m2 * m2 * R2 * R2 + 2 * m1 * m2 * R1 * R1 * R2 * R2 +
                2 * I1 * (I2 + m2 * R2 * R2) - L1 * L1 * m2 * m2 * R2 * R2 * Math.cos(2 * (th1 - th2 + phi))));
        change[THETA2] = dth2;
        change[THETA2P] = (m2 * R2 * (-(g * (2 * I1 + L1 * L1 * m2 + 2 * m1 * R1 * R1) * Math.sin(th2)) +
            L1 * (g * m1 * R1 * Math.sin(th2 - phi) +
                2 * dth1 * dth1 * (I1 + L1 * L1 * m2 + m1 * R1 * R1) * Math.sin(th1 - th2 + phi) +
                dth2 * dth2 * L1 * m2 * R2 * Math.sin(2 * (th1 - th2 + phi)) +
                g * m1 * R1 * Math.sin(2 * th1 - th2 + phi) +
                g * L1 * m2 * Math.sin(2 * th1 - th2 + 2 * phi)))) /
            (2 * I2 * L1 * L1 * m2 + 2 * I2 * m1 * R1 * R1 + L1 * L1 * m2 * m2 * R2 * R2 +
                2 * m1 * m2 * R1 * R1 * R2 * R2 + 2 * I1 * (I2 + m2 * R2 * R2) -
                L1 * L1 * m2 * m2 * R2 * R2 * Math.cos(2 * (th1 - th2 + phi)));
        return null;
    }
    ;
    getGamma1() {
        return this.gamma1_;
    }
    ;
    getGamma2() {
        return this.gamma2_;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RigidDoublePendulumSim.en.GRAVITY);
    }
    ;
    setInitialAngles() {
        this.reset();
        const va = this.getVarsList();
        const vars = va.getValues();
        vars[THETA1] = this.omega1_ + this.gamma1_;
        vars[THETA1P] = 0;
        vars[THETA2] = this.omega2_ + this.gamma2_;
        vars[THETA2P] = 0;
        va.setValues(vars);
        this.saveInitialState();
        this.modifyObjects();
    }
    ;
    getAngle1() {
        return this.omega1_;
    }
    ;
    setAngle1(value) {
        this.omega1_ = value;
        this.setInitialAngles();
        this.broadcastParameter(RigidDoublePendulumSim.en.ANGLE_1);
    }
    ;
    getAngle2() {
        return this.omega2_;
    }
    ;
    setAngle2(value) {
        this.omega2_ = value;
        this.setInitialAngles();
        this.broadcastParameter(RigidDoublePendulumSim.en.ANGLE_2);
    }
    ;
}
RigidDoublePendulumSim.en = {
    ANGLE_1: 'angle-1',
    ANGLE_1_VELOCITY: 'angle-1 velocity',
    ANGLE_2: 'angle-2',
    ANGLE_2_VELOCITY: 'angle-2 velocity',
    GRAVITY: 'gravity',
    MASS_1: 'mass-1',
    MASS_2: 'mass-2',
    ROD_1_LENGTH: 'rod-1 length',
    ROD_2_LENGTH: 'rod-2 length',
    PENDULUM: 'pendulum'
};
RigidDoublePendulumSim.de_strings = {
    ANGLE_1: 'Winkel-1',
    ANGLE_1_VELOCITY: 'Winkel-1 Geschwindigkeit',
    ANGLE_2: 'Winkel-2',
    ANGLE_2_VELOCITY: 'Winkel-2 Geschwindigkeit',
    GRAVITY: 'Gravitation',
    MASS_1: 'Masse-1',
    MASS_2: 'Masse-2',
    ROD_1_LENGTH: 'Stange-1 Länge',
    ROD_2_LENGTH: 'Stange-2 Länge',
    PENDULUM: 'Pendel'
};
RigidDoublePendulumSim.i18n = Util.LOCALE === 'de' ? RigidDoublePendulumSim.de_strings : RigidDoublePendulumSim.en;
Util.defineGlobal('sims$pendulum$RigidDoublePendulumSim', RigidDoublePendulumSim);
