import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { HumpPath } from './HumpPath.js';
import { NumericalPath } from '../../lab/model/NumericalPath.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const X = 0;
const XP = 1;
const S = 2;
const SP = 3;
const KE = 4;
const PE = 5;
const TE = 6;
const TIME = 7;
const Y = 8;
const YP = 9;
export class LagrangeRollerSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.gravity_ = 2;
        this.potentialOffset_ = 0;
        this.dragObj_ = null;
        const var_names = [
            LagrangeRollerSim.en.X_POSITION,
            LagrangeRollerSim.en.X_VELOCITY,
            LagrangeRollerSim.en.POSITION,
            LagrangeRollerSim.en.VELOCITY,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME,
            LagrangeRollerSim.en.Y_POSITION,
            LagrangeRollerSim.en.Y_VELOCITY
        ];
        const i18n_names = [
            LagrangeRollerSim.i18n.X_POSITION,
            LagrangeRollerSim.i18n.X_VELOCITY,
            LagrangeRollerSim.i18n.POSITION,
            LagrangeRollerSim.i18n.VELOCITY,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME,
            LagrangeRollerSim.i18n.Y_POSITION,
            LagrangeRollerSim.i18n.Y_VELOCITY
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(S, SP, KE, PE, TE, Y, YP);
        this.ball1_ = PointMass.makeCircle(0.3, 'ball1');
        this.ball1_.setMass(0.5);
        this.getSimList().add(this.ball1_);
        this.path_ = new NumericalPath(new HumpPath());
        this.lowestPoint_ = this.path_.getBoundsWorld().getBottom();
        const va = this.getVarsList();
        const vars = va.getValues();
        vars[X] = 3;
        vars[XP] = 0;
        vars[S] = this.path_.map_x_to_p(vars[X]);
        va.setValues(vars);
        this.modifyObjects();
        this.saveInitialState();
        this.addParameter(new ParameterNumber(this, LagrangeRollerSim.en.GRAVITY, LagrangeRollerSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, LagrangeRollerSim.en.MASS, LagrangeRollerSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        let pn;
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', ball1_: ' + this.ball1_
            + ', path_: ' + this.path_
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', lowestPoint_: ' + Util.NF(this.lowestPoint_)
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'LagrangeRollerSim';
    }
    ;
    getPath() {
        return this.path_;
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues(false);
        this.moveObjects(vars);
        const x = vars[X];
        vars[S] = this.path_.map_x_to_p(x);
        const d = -(7 / 3) * x + (2 / 3) * x * x * x;
        vars[SP] = Math.sqrt(1 + d * d) * vars[XP];
        const ei = this.getEnergyInfo_(vars);
        vars[KE] = ei.getTranslational();
        vars[PE] = ei.getPotential();
        vars[TE] = ei.getTotalEnergy();
        vars[Y] = this.ball1_.getPosition().getY();
        vars[YP] = this.ball1_.getVelocity().getY();
        this.getVarsList().setValues(vars, true);
    }
    ;
    moveObjects(vars) {
        const x = vars[X];
        const x2 = x * x;
        const x3 = x2 * x;
        const x4 = x3 * x;
        const y = 3 - (7 / 6) * x2 + (1 / 6) * x4;
        this.ball1_.setPosition(new Vector(x, y));
        const yp = (-(7 / 3) * x + (2 / 3) * x3) * vars[XP];
        this.ball1_.setVelocity(new Vector(vars[XP], yp));
    }
    ;
    getEnergyInfo() {
        const vars = this.getVarsList().getValues(true);
        return this.getEnergyInfo_(vars);
    }
    ;
    getEnergyInfo_(vars) {
        const ke = 0.5 * this.ball1_.getMass() * vars[SP] * vars[SP];
        const ke2 = this.ball1_.getKineticEnergy();
        if (Util.veryDifferent(ke, ke2)) {
            throw 'kinetic energy calcs differ ' + ke + ' vs ' + ke2;
        }
        const pe = this.ball1_.getMass() * this.gravity_ *
            (this.ball1_.getPosition().getY() - this.lowestPoint_);
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.ball1_) {
            this.dragObj_ = simObject;
            return true;
        }
        return false;
    }
    ;
    mouseDrag(simObject, location, offset) {
        const p = location.subtract(offset);
        if (simObject == this.ball1_) {
            const pathPoint = this.path_.findNearestGlobal(p);
            const va = this.getVarsList();
            const vars = va.getValues();
            vars[X] = pathPoint.x;
            vars[XP] = 0;
            vars[S] = pathPoint.p;
            va.setValues(vars);
            this.moveObjects(vars);
        }
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragObj_ = null;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        if (this.dragObj_ != this.ball1_) {
            const x = vars[X];
            const x2 = x * x;
            const v = vars[XP];
            change[X] = v;
            const r = -(x * (-7 + 2 * x2) * (3 * this.gravity_ + (-7 + 6 * x2) * v * v));
            change[XP] = r / (9 + 49 * x2 - 28 * x2 * x2 + 4 * x2 * x2 * x2);
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(LagrangeRollerSim.en.GRAVITY);
    }
    ;
    getMass() {
        return this.ball1_.getMass();
    }
    setMass(value) {
        this.ball1_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(LagrangeRollerSim.en.MASS);
    }
}
LagrangeRollerSim.en = {
    GRAVITY: 'gravity',
    MASS: 'mass',
    POSITION: 'position',
    VELOCITY: 'velocity',
    X_POSITION: 'X position',
    Y_POSITION: 'Y position',
    X_VELOCITY: 'X velocity',
    Y_VELOCITY: 'Y velocity'
};
LagrangeRollerSim.de_strings = {
    GRAVITY: 'Gravitation',
    MASS: 'Masse',
    POSITION: 'Position',
    VELOCITY: 'Geschwindigkeit',
    X_POSITION: 'X Position',
    Y_POSITION: 'Y Position',
    X_VELOCITY: 'X Geschwindigkeit',
    Y_VELOCITY: 'Y Geschwindigkeit'
};
LagrangeRollerSim.i18n = Util.LOCALE === 'de' ? LagrangeRollerSim.de_strings : LagrangeRollerSim.en;
Util.defineGlobal('sims$roller$LagrangeRollerSim', LagrangeRollerSim);
