import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PathPoint } from '../../lab/model/PathPoint.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const P1 = 0;
const V1 = 1;
const P2 = 2;
const V2 = 3;
const KE = 4;
const PE = 5;
const TE = 6;
const TIME = 7;
export class RollerDoubleSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.damping_ = 0.001;
        this.gravity_ = 9.8;
        this.path_ = null;
        this.lowestPoint_ = 0;
        this.potentialOffset_ = 0;
        this.pathPoint1_ = new PathPoint();
        this.pathPoint2_ = new PathPoint();
        this.dragObj_ = null;
        const var_names = [
            RollerDoubleSim.en.POSITION_1,
            RollerDoubleSim.en.VELOCITY_1,
            RollerDoubleSim.en.POSITION_2,
            RollerDoubleSim.en.VELOCITY_2,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME
        ];
        const i18n_names = [
            RollerDoubleSim.i18n.POSITION_1,
            RollerDoubleSim.i18n.VELOCITY_1,
            RollerDoubleSim.i18n.POSITION_2,
            RollerDoubleSim.i18n.VELOCITY_2,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.ball1_ = PointMass.makeCircle(0.3, 'ball1');
        this.ball1_.setMass(0.5);
        this.ball2_ = PointMass.makeCircle(0.3, 'ball2');
        this.ball2_.setMass(0.5);
        this.spring_ = new Spring('spring', this.ball1_, Vector.ORIGIN, this.ball2_, Vector.ORIGIN, 2.0, 6.0);
        this.getSimList().add(this.ball1_, this.ball2_, this.spring_);
        let pn;
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.DAMPING, RollerDoubleSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.GRAVITY, RollerDoubleSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.MASS_1, RollerDoubleSim.i18n.MASS_1, () => this.getMass1(), a => this.setMass1(a)));
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.MASS_2, RollerDoubleSim.i18n.MASS_2, () => this.getMass2(), a => this.setMass2(a)));
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.SPRING_DAMPING, RollerDoubleSim.i18n.SPRING_DAMPING, () => this.getSpringDamping(), a => this.setSpringDamping(a)));
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.SPRING_LENGTH, RollerDoubleSim.i18n.SPRING_LENGTH, () => this.getSpringLength(), a => this.setSpringLength(a)));
        this.addParameter(new ParameterNumber(this, RollerDoubleSim.en.SPRING_STIFFNESS, RollerDoubleSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', ball1_: ' + this.ball1_
            + ', ball2_: ' + this.ball2_
            + ', spring_: ' + this.spring_
            + ', path_: ' + this.path_
            + ', damping_: ' + Util.NF(this.damping_)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', lowestPoint_: ' + Util.NF(this.lowestPoint_)
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'RollerDoubleSim';
    }
    ;
    getPath() {
        return this.path_;
    }
    ;
    setPath(path) {
        const simList = this.getSimList();
        const oldPath = this.path_;
        if (oldPath != null) {
            simList.remove(oldPath);
        }
        this.path_ = path;
        simList.add(path);
        const r = path.getBoundsWorld();
        this.lowestPoint_ = r.getBottom();
        const start1 = new Vector(r.getLeft() + r.getWidth() * 0.1, r.getTop() - r.getHeight() * 0.1);
        this.pathPoint1_ = path.findNearestGlobal(start1);
        const start2 = new Vector(r.getLeft() + r.getWidth() * 0.2, r.getTop() - r.getHeight() * 0.3);
        this.pathPoint2_ = path.findNearestGlobal(start2);
        this.getVarsList().setValues([this.pathPoint1_.p, 0, this.pathPoint2_.p, 0]);
        this.saveInitialState();
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        if (this.path_ != null) {
            const p = this.path_.mod_p(vars[P1]);
            if (p != vars[P1]) {
                vars[P1] = p;
                va.setValue(P1, p);
            }
            const p2 = this.path_.mod_p(vars[P2]);
            if (p2 != vars[P2]) {
                vars[P2] = p2;
                va.setValue(P2, p2);
            }
            const ei = this.getEnergyInfo();
            va.setValue(KE, ei.getTranslational(), true);
            va.setValue(PE, ei.getPotential(), true);
            va.setValue(TE, ei.getTotalEnergy(), true);
        }
    }
    ;
    moveObjects(vars) {
        if (this.path_ != null) {
            const pp1 = this.pathPoint1_;
            const pp2 = this.pathPoint2_;
            pp1.p = vars[P1];
            pp2.p = vars[P2];
            this.path_.map_p_to_slope(pp1);
            this.path_.map_p_to_slope(pp2);
            this.ball1_.setPosition(pp1);
            this.ball2_.setPosition(pp2);
            this.ball1_.setVelocity(pp1.getSlope().multiply(vars[V1]));
            this.ball2_.setVelocity(pp2.getSlope().multiply(vars[V2]));
        }
    }
    ;
    getEnergyInfo() {
        const ke = this.ball1_.getKineticEnergy() + this.ball2_.getKineticEnergy();
        let pe = this.ball1_.getMass() * this.gravity_
            * (this.ball1_.getPosition().getY() - this.lowestPoint_);
        pe += this.ball2_.getMass() * this.gravity_
            * (this.ball2_.getPosition().getY() - this.lowestPoint_);
        pe += this.spring_.getPotentialEnergy();
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.ball1_) {
            this.dragObj_ = simObject;
            return true;
        }
        else if (simObject == this.ball2_) {
            this.dragObj_ = simObject;
            return true;
        }
        return false;
    }
    ;
    mouseDrag(simObject, location, offset) {
        if (this.path_ == null)
            return;
        const va = this.getVarsList();
        const p = location.subtract(offset);
        if (simObject == this.ball1_) {
            this.pathPoint1_ = this.path_.findNearestGlobal(p);
            va.setValue(P1, this.pathPoint1_.p);
            va.setValue(V1, 0);
            va.incrSequence(KE, PE, TE);
        }
        else if (simObject == this.ball2_) {
            this.pathPoint2_ = this.path_.findNearestGlobal(p);
            va.setValue(P2, this.pathPoint2_.p);
            va.setValue(V2, 0);
            va.incrSequence(KE, PE, TE);
        }
        this.moveObjects(va.getValues());
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragObj_ = null;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        this.moveObjects(vars);
        const springForces = this.spring_.calculateForces();
        if (this.dragObj_ != this.ball1_) {
            change[P1] = vars[V1];
            const k = this.pathPoint1_.slope;
            const sinTheta = isFinite(k) ? k / Math.sqrt(1 + k * k) : 1;
            change[V1] = -this.gravity_ * this.pathPoint1_.direction * sinTheta
                - this.damping_ * vars[V1] / this.ball1_.getMass();
            let tangent;
            if (!isFinite(k)) {
                tangent = new Vector(0, k > 0 ? 1 : -1, 0);
            }
            else {
                tangent = new Vector(1, k, 0);
                tangent = tangent.normalize().multiply(this.pathPoint1_.direction);
            }
            const force = springForces[0];
            Util.assert(force.getBody() == this.ball1_);
            const f = force.getVector();
            change[V1] += f.dotProduct(tangent) / this.ball1_.getMass();
        }
        if (this.dragObj_ != this.ball2_) {
            change[P2] = vars[V2];
            const k = this.pathPoint2_.slope;
            const sinTheta = isFinite(k) ? k / Math.sqrt(1 + k * k) : 1;
            change[V2] = -this.gravity_ * this.pathPoint2_.direction * sinTheta
                - this.damping_ * vars[V2] / this.ball2_.getMass();
            let tangent;
            if (!isFinite(k)) {
                tangent = new Vector(0, k > 0 ? 1 : -1, 0);
            }
            else {
                tangent = new Vector(1, k, 0);
                tangent = tangent.normalize().multiply(this.pathPoint2_.direction);
            }
            const force = springForces[1];
            Util.assert(force.getBody() == this.ball2_);
            const f = force.getVector();
            change[V2] += f.dotProduct(tangent) / this.ball2_.getMass();
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerDoubleSim.en.GRAVITY);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(RollerDoubleSim.en.DAMPING);
    }
    getMass1() {
        return this.ball1_.getMass();
    }
    setMass1(value) {
        this.ball1_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(RollerDoubleSim.en.MASS_1);
    }
    getMass2() {
        return this.ball2_.getMass();
    }
    setMass2(value) {
        this.ball2_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(RollerDoubleSim.en.MASS_2);
    }
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerDoubleSim.en.SPRING_STIFFNESS);
    }
    getSpringLength() {
        return this.spring_.getRestLength();
    }
    setSpringLength(value) {
        this.spring_.setRestLength(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerDoubleSim.en.SPRING_LENGTH);
    }
    getSpringDamping() {
        return this.spring_.getDamping();
    }
    setSpringDamping(value) {
        this.spring_.setDamping(value);
        this.broadcastParameter(RollerDoubleSim.en.SPRING_DAMPING);
    }
}
RollerDoubleSim.en = {
    DAMPING: 'damping',
    GRAVITY: 'gravity',
    MASS_1: 'mass-1',
    MASS_2: 'mass-2',
    POSITION_1: 'position-1',
    POSITION_2: 'position-2',
    SPRING_DAMPING: 'spring damping',
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    VELOCITY_1: 'velocity-1',
    VELOCITY_2: 'velocity-2'
};
RollerDoubleSim.de_strings = {
    DAMPING: 'Dämpfung',
    GRAVITY: 'Gravitation',
    MASS_1: 'Masse-1',
    MASS_2: 'Masse-2',
    POSITION_1: 'Position-1',
    POSITION_2: 'Position-2',
    SPRING_DAMPING: 'Federdämpfung',
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    VELOCITY_1: 'Geschwindigkeit-1',
    VELOCITY_2: 'Geschwindigkeit-2'
};
RollerDoubleSim.i18n = Util.LOCALE === 'de' ? RollerDoubleSim.de_strings : RollerDoubleSim.en;
Util.defineGlobal('sims$roller$RollerDoubleSim', RollerDoubleSim);
