import { AbstractODESim } from '../../lab/model/ODESim.js';
import { BlockCollision } from './BlockCollision.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const X1 = 0;
const V1 = 1;
const X2 = 2;
const V2 = 3;
const TIME = 4;
const KE = 5;
const PE = 6;
const TE = 7;
export class CollideBlocksSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.damping_ = 0;
        this.potentialOffset_ = 0;
        this.debugPaint_ = null;
        this.isDragging = false;
        this.collisionFunction_ = null;
        const var_names = [
            CollideBlocksSim.en.POSITION_1,
            CollideBlocksSim.en.VELOCITY_1,
            CollideBlocksSim.en.POSITION_2,
            CollideBlocksSim.en.VELOCITY_2,
            VarsList.en.TIME,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY
        ];
        const i18n_names = [
            CollideBlocksSim.i18n.POSITION_1,
            CollideBlocksSim.i18n.VELOCITY_1,
            CollideBlocksSim.i18n.POSITION_2,
            CollideBlocksSim.i18n.VELOCITY_2,
            VarsList.i18n.TIME,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_SIM'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.wallLeft_ = PointMass.makeRectangle(0.4, 4, 'wallLeft');
        this.wallLeft_.setMass(Infinity);
        this.wallLeft_.setPosition(new Vector(-0.2, 0));
        this.wallRight_ = PointMass.makeRectangle(0.4, 4, 'wallRight');
        this.wallRight_.setMass(Infinity);
        this.wallRight_.setPosition(new Vector(7.2, 0));
        this.block1_ = PointMass.makeSquare(0.6, 'block1');
        this.block1_.setMass(0.5);
        this.block2_ = PointMass.makeSquare(0.6, 'block2');
        this.block2_.setMass(1.5);
        this.spring1_ = new Spring('spring1', this.wallLeft_, new Vector(this.wallLeft_.getWidth() / 2, 0), this.block1_, Vector.ORIGIN, 2.5, 6.0);
        this.spring2_ = new Spring('spring2', this.wallRight_, new Vector(-this.wallRight_.getWidth() / 2, 0), this.block2_, Vector.ORIGIN, 2.5, 0);
        this.getVarsList().setValues([0.5, 0, 3, 0]);
        this.saveInitialState();
        let pn;
        this.getSimList().add(this.wallLeft_, this.wallRight_, this.spring1_, this.spring2_, this.block1_, this.block2_);
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.MASS_1, CollideBlocksSim.i18n.MASS_1, () => this.getMass1(), a => this.setMass1(a)));
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.MASS_2, CollideBlocksSim.i18n.MASS_2, () => this.getMass2(), a => this.setMass2(a)));
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.LENGTH_1, CollideBlocksSim.i18n.LENGTH_1, () => this.getSpring1Length(), a => this.setSpring1Length(a)));
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.STIFFNESS_1, CollideBlocksSim.i18n.STIFFNESS_1, () => this.getSpring1Stiffness(), a => this.setSpring1Stiffness(a)));
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.LENGTH_2, CollideBlocksSim.i18n.LENGTH_2, () => this.getSpring2Length(), a => this.setSpring2Length(a)));
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.STIFFNESS_2, CollideBlocksSim.i18n.STIFFNESS_2, () => this.getSpring2Stiffness(), a => this.setSpring2Stiffness(a)));
        this.addParameter(new ParameterNumber(this, CollideBlocksSim.en.DAMPING, CollideBlocksSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', block1_: ' + this.block1_
            + ', block2_: ' + this.block2_
            + ', wallLeft_: ' + this.wallLeft_
            + ', wallRight_: ' + this.wallRight_
            + ', spring1_: ' + this.spring1_
            + ', spring2_: ' + this.spring2_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'CollideBlocksSim';
    }
    ;
    getEnergyInfo() {
        const ke = this.block1_.getKineticEnergy() + this.block2_.getKineticEnergy();
        const pe = this.spring1_.getPotentialEnergy() + this.spring2_.getPotentialEnergy();
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        vars[KE] = ei.getTranslational();
        vars[PE] = ei.getPotential();
        vars[TE] = ei.getTotalEnergy();
        va.setValues(vars, true);
    }
    ;
    moveObjects(vars) {
        this.block1_.setPosition(new Vector(vars[X1], 0));
        this.block1_.setVelocity(new Vector(vars[V1], 0));
        this.block2_.setPosition(new Vector(vars[X2], 0));
        this.block2_.setVelocity(new Vector(vars[V2], 0));
        if (this.debugPaint_ != null) {
            this.debugPaint_();
        }
    }
    ;
    setDebugPaint(fn) {
        this.debugPaint_ = fn;
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.block1_ || simObject == this.block2_) {
            this.isDragging = true;
            return true;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const gap = 0.02;
        const va = this.getVarsList();
        const vars = va.getValues();
        const p = location.subtract(offset);
        if (simObject == this.block1_) {
            let x = p.getX();
            if (x - this.block1_.getWidth() / 2 < this.wallLeft_.getRightWorld() + gap) {
                x = this.wallLeft_.getRightWorld() + this.block1_.getWidth() / 2 + gap;
            }
            if (x + this.block1_.getWidth() / 2 + this.block2_.getWidth() >
                this.wallRight_.getLeftWorld() - 2 * gap) {
                x = this.wallRight_.getLeftWorld() - this.block2_.getWidth() -
                    this.block1_.getWidth() / 2 - 2 * gap;
            }
            if (x + this.block1_.getWidth() / 2 > this.block2_.getLeftWorld() - gap) {
                this.block2_.setPosition(new Vector(x + this.block1_.getWidth() / 2 +
                    this.block2_.getWidth() / 2 + gap, 0));
            }
            vars[X1] = x;
            vars[V1] = 0;
            vars[X2] = this.block2_.getPosition().getX();
            vars[V2] = 0;
            va.setValues(vars);
        }
        else if (simObject == this.block2_) {
            let x = p.getX();
            if (x + this.block2_.getWidth() / 2 > this.wallRight_.getLeftWorld() - gap) {
                x = this.wallRight_.getLeftWorld() - this.block2_.getWidth() / 2 - gap;
            }
            if (x - this.block2_.getWidth() / 2 - this.block1_.getWidth() <
                this.wallLeft_.getRightWorld() + 2 * gap) {
                x = this.wallLeft_.getRightWorld() + this.block1_.getWidth() +
                    this.block2_.getWidth() / 2 + 2 * gap;
            }
            if (x - this.block2_.getWidth() / 2 < this.block1_.getRightWorld() + gap) {
                this.block1_.setPosition(new Vector(x - this.block2_.getWidth() / 2 -
                    this.block1_.getWidth() / 2 - gap, 0));
            }
            vars[X1] = this.block1_.getPosition().getX();
            vars[V1] = 0;
            vars[X2] = x;
            vars[V2] = 0;
            va.setValues(vars);
        }
        this.moveObjects(vars);
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging = false;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    addCollision(collisions, leftBlock, rightBlock, time) {
        const c = new BlockCollision(leftBlock, rightBlock, time);
        if (c.getDistance() < 0.1) {
            const similar = collisions.find(c2 => c.similarTo(c2));
            if (similar) {
                const c2 = similar;
                if (!c2.needsHandling() && c.getDistance() < c2.getDistance()) {
                    Util.remove(collisions, c2);
                    collisions.push(c);
                }
            }
            else {
                collisions.push(c);
            }
        }
    }
    ;
    findCollisions(collisions, vars, stepSize) {
        this.moveObjects(vars);
        const time = this.getTime() + stepSize;
        this.addCollision(collisions, this.wallLeft_, this.block1_, time);
        this.addCollision(collisions, this.block1_, this.block2_, time);
        this.addCollision(collisions, this.block2_, this.wallRight_, time);
    }
    ;
    handleCollisions(collisions, opt_totals) {
        const va = this.getVarsList();
        const seq0 = va.getVariable(0).getSequence();
        const seq2 = va.getVariable(2).getSequence();
        const vars = va.getValues();
        collisions.forEach(c => {
            if (c.leftBlock_ == this.wallLeft_ && c.rightBlock_ == this.block1_) {
                c.impulse = Math.abs(vars[V1] * this.block1_.getMass());
                va.setValue(V1, -vars[V1]);
            }
            else if (c.rightBlock_ == this.wallRight_ && c.leftBlock_ == this.block2_) {
                c.impulse = Math.abs(vars[V2] * this.block2_.getMass());
                va.setValue(V2, -vars[V2]);
            }
            else if (c.leftBlock_ == this.block1_ && c.rightBlock_ == this.block2_) {
                const vcm = (this.block1_.getMass() * vars[V1] + this.block2_.getMass() * vars[V2]) /
                    (this.block1_.getMass() + this.block2_.getMass());
                c.impulse = Math.abs(2 * (vcm - vars[V2]) * this.block2_.getMass());
                va.setValue(V1, -vars[V1] + 2 * vcm);
                va.setValue(V2, -vars[V2] + 2 * vcm);
            }
            else {
                throw 'illegal collision ' + c;
            }
            if (opt_totals) {
                opt_totals.addImpulses(1);
            }
            if (this.collisionFunction_ && this.terminal_) {
                this.collisionFunction_(c, this.terminal_);
            }
        });
        Util.assert(va.getVariable(X1).getSequence() == seq0);
        Util.assert(va.getVariable(X2).getSequence() == seq2);
        return true;
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1.0;
        this.moveObjects(vars);
        if (!this.isDragging) {
            change[X1] = vars[V1];
            change[V1] = (-this.spring1_.getStiffness() * this.spring1_.getStretch() -
                this.damping_ * vars[V1]) / this.block1_.getMass();
            change[X2] = vars[V2];
            change[V2] = (this.spring2_.getStiffness() * this.spring2_.getStretch() -
                this.damping_ * vars[V2]) / this.block2_.getMass();
        }
        return null;
    }
    ;
    getMomentum() {
        this.modifyObjects();
        const m1 = this.block1_.getVelocity().multiply(this.block1_.getMass());
        const m2 = this.block2_.getVelocity().multiply(this.block2_.getMass());
        return m1.add(m2).length();
    }
    ;
    getMass1() {
        return this.block1_.getMass();
    }
    ;
    setMass1(value) {
        this.block1_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(CollideBlocksSim.en.MASS_1);
    }
    ;
    getMass2() {
        return this.block2_.getMass();
    }
    ;
    setMass2(value) {
        this.block2_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(CollideBlocksSim.en.MASS_2);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(CollideBlocksSim.en.DAMPING);
    }
    ;
    getSpring1Stiffness() {
        return this.spring1_.getStiffness();
    }
    ;
    setSpring1Stiffness(value) {
        this.spring1_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(CollideBlocksSim.en.STIFFNESS_1);
    }
    ;
    getSpring1Length() {
        return this.spring1_.getRestLength();
    }
    ;
    setSpring1Length(value) {
        this.spring1_.setRestLength(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(CollideBlocksSim.en.LENGTH_1);
    }
    ;
    getSpring2Stiffness() {
        return this.spring2_.getStiffness();
    }
    ;
    setSpring2Stiffness(value) {
        this.spring2_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(CollideBlocksSim.en.STIFFNESS_2);
    }
    ;
    getSpring2Length() {
        return this.spring2_.getRestLength();
    }
    ;
    setSpring2Length(value) {
        this.spring2_.setRestLength(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(CollideBlocksSim.en.LENGTH_2);
    }
    ;
    setCollisionFunction(f) {
        this.collisionFunction_ = f;
    }
    ;
}
CollideBlocksSim.en = {
    ACCELERATION: 'acceleration',
    DAMPING: 'damping',
    MASS_1: 'mass 1',
    MASS_2: 'mass 2',
    POSITION_1: 'position 1',
    POSITION_2: 'position 2',
    LENGTH_1: 'spring 1 length',
    STIFFNESS_1: 'spring 1 stiffness',
    LENGTH_2: 'spring 2 length',
    STIFFNESS_2: 'spring 2 stiffness',
    VELOCITY_1: 'velocity 1',
    VELOCITY_2: 'velocity 2'
};
CollideBlocksSim.de_strings = {
    ACCELERATION: 'Beschleunigung',
    DAMPING: 'Dämpfung',
    MASS_1: 'Masse 1',
    MASS_2: 'Masse 2',
    POSITION_1: 'Position 1',
    POSITION_2: 'Position 2',
    LENGTH_1: 'Feder 1 Länge',
    STIFFNESS_1: 'Feder 1 Steifigkeit',
    LENGTH_2: 'Feder 2 Länge',
    STIFFNESS_2: 'Feder 2 Steifigkeit',
    VELOCITY_1: 'Geschwindigkeit 1',
    VELOCITY_2: 'Geschwindigkeit 2'
};
CollideBlocksSim.i18n = Util.LOCALE === 'de' ? CollideBlocksSim.de_strings : CollideBlocksSim.en;
Util.defineGlobal('sims$springs$CollideBlocksSim', CollideBlocksSim);
