import { AbstractApp } from '../common/AbstractApp.js';
import { ChoiceControl } from '../../lab/controls/ChoiceControl.js';
import { CollideSpringSim } from './CollideSpringSim.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DisplaySpring } from '../../lab/view/DisplaySpring.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { SimList } from '../../lab/model/SimList.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
export class CollideSpringApp extends AbstractApp {
    constructor(elem_ids) {
        Util.setErrorHandler();
        const simRect = new DoubleRect(-6.4, -2, 6.4, 2);
        const sim = new CollideSpringSim();
        const advance = new SimpleAdvance(sim);
        super(elem_ids, simRect, sim, advance, sim, sim);
        this.startPosition = CollideSpringSim.START_MIDDLE;
        this.numBlocks = 3;
        this.startGap = 0.1;
        this.protoBlock = new DisplayShape();
        this.protoWall = new DisplayShape();
        this.protoSpring = new DisplaySpring();
        this.layout.getSimCanvas().setBackground('black');
        this.protoBlock.setFillStyle('blue');
        this.protoWall.setFillStyle('lightGray');
        this.protoSpring.setWidth(0.3);
        this.protoSpring.setColorCompressed('#0c0');
        this.protoSpring.setColorExpanded('#6f6');
        Util.assert(this.simList.length() == 0);
        this.simList.addObserver(this);
        this.sim.config(this.numBlocks, this.startPosition, this.startGap);
        this.simRun.setTimeStep(0.0025);
        this.addPlaybackControls();
        let pn;
        this.addParameter(pn = new ParameterNumber(this, CollideSpringSim.en.NUM_BLOCKS, CollideSpringSim.i18n.NUM_BLOCKS, () => this.getNumBlocks(), a => this.setNumBlocks(a)));
        pn.setDecimalPlaces(0);
        pn.setLowerLimit(1);
        pn.setUpperLimit(3);
        this.addControl(new SliderControl(pn, 1, 3, false));
        this.addParameter(pn = new ParameterNumber(this, CollideSpringSim.en.START_POSITION, CollideSpringSim.i18n.START_POSITION, () => this.getStartPosition(), a => this.setStartPosition(a), [CollideSpringSim.i18n.START_IN_MIDDLE,
            CollideSpringSim.i18n.START_ON_WALL], [CollideSpringSim.START_MIDDLE, CollideSpringSim.START_ON_WALL]));
        this.addControl(new ChoiceControl(pn));
        this.addParameter(pn = new ParameterNumber(this, CollideSpringSim.en.START_GAP, CollideSpringSim.i18n.START_GAP, () => this.getStartGap(), a => this.setStartGap(a)));
        this.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(CollideSpringSim.en.DAMPING);
        this.addControl(new SliderControl(pn, 0, 1, false));
        pn = sim.getParameterNumber(CollideSpringSim.en.SPRING_DAMPING);
        this.addControl(new SliderControl(pn, 0, 1, false));
        pn = sim.getParameterNumber(CollideSpringSim.en.MASS);
        this.addControl(new SliderControl(pn, 0.2, 50.2, true));
        pn = sim.getParameterNumber(CollideSpringSim.en.SPRING_LENGTH);
        this.addControl(new SliderControl(pn, 0, 1.5, false));
        pn = sim.getParameterNumber(CollideSpringSim.en.SPRING_STIFFNESS);
        this.addControl(new SliderControl(pn, 0, 100, false));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        pn = this.timeGraph.line1.getParameterNumber('y-variable');
        pn.setValue(1);
        pn = this.timeGraph.line2.getParameterNumber('y-variable');
        pn.setValue(2);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', startGap: ' + Util.NF(this.startGap)
            + ', numBlocks: ' + Util.NF(this.numBlocks)
            + ', startPosition: ' + Util.NF(this.startPosition)
            + super.toString();
    }
    ;
    getClassName() {
        return 'CollideSpringApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('protoBlock|protoWall|protoSpring', myName + '.');
    }
    ;
    addBody(obj) {
        if (this.displayList.find(obj)) {
            return;
        }
        if (obj instanceof PointMass) {
            const rm = obj;
            const proto = rm.getName().match(/^WALL/) ? this.protoWall : this.protoBlock;
            this.displayList.add(new DisplayShape(rm, proto));
        }
        else if (obj instanceof Spring) {
            const s = obj;
            this.displayList.add(new DisplaySpring(s, this.protoSpring));
        }
    }
    ;
    observe(event) {
        if (event.getSubject() == this.simList) {
            const obj = event.getValue();
            if (event.nameEquals(SimList.OBJECT_ADDED)) {
                this.addBody(obj);
            }
            else if (event.nameEquals(SimList.OBJECT_REMOVED)) {
                const d = this.displayList.find(obj);
                if (d) {
                    this.displayList.remove(d);
                }
            }
        }
    }
    ;
    getStartGap() {
        return this.startGap;
    }
    ;
    setStartGap(value) {
        if (value != this.startGap) {
            this.startGap = value;
            this.sim.config(this.numBlocks, this.startPosition, this.startGap);
            this.broadcastParameter(CollideSpringSim.en.START_GAP);
            this.easyScript.update();
        }
    }
    ;
    getNumBlocks() {
        return this.numBlocks;
    }
    ;
    setNumBlocks(value) {
        value = Math.floor(value + 0.5);
        if (this.numBlocks != value) {
            this.numBlocks = value;
            this.sim.config(this.numBlocks, this.startPosition, this.startGap);
            this.broadcastParameter(CollideSpringSim.en.NUM_BLOCKS);
            this.easyScript.update();
        }
    }
    ;
    getStartPosition() {
        return this.startPosition;
    }
    ;
    setStartPosition(value) {
        if (this.startPosition != value) {
            this.startPosition = value;
            this.sim.config(this.numBlocks, this.startPosition, this.startGap);
            this.broadcastParameter(CollideSpringSim.en.START_POSITION);
            this.easyScript.update();
        }
    }
    ;
}
Util.defineGlobal('sims$springs$CollideSpringApp', CollideSpringApp);
