import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const U1X = 0;
const U1Y = 1;
const U2X = 2;
const U2Y = 3;
const V1X = 4;
const V1Y = 5;
const V2X = 6;
const V2Y = 7;
const KE = 8;
const PE = 9;
const TE = 10;
const TIME = 11;
const FIX_X = 12;
const FIX_Y = 13;
export class Double2DSpringSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.dragBlock_ = -1;
        this.gravity_ = 9.8;
        this.damping_ = 0;
        this.potentialOffset_ = 0;
        this.topMass_ = PointMass.makeSquare(0.5, 'top');
        this.bob1_ = PointMass.makeCircle(0.5, 'bob1');
        this.bob2_ = PointMass.makeCircle(0.5, 'bob2');
        this.spring1_ = new Spring('spring1', this.topMass_, Vector.ORIGIN, this.bob1_, Vector.ORIGIN, 1.0, 6.0);
        this.spring2_ = new Spring('spring2', this.bob1_, Vector.ORIGIN, this.bob2_, Vector.ORIGIN, 1.0, 6.0);
        this.springs_ = [this.spring1_, this.spring2_];
        const var_names = [
            Double2DSpringSim.en.X_POSITION + '-1',
            Double2DSpringSim.en.Y_POSITION + '-1',
            Double2DSpringSim.en.X_POSITION + '-2',
            Double2DSpringSim.en.Y_POSITION + '-2',
            Double2DSpringSim.en.X_VELOCITY + '-1',
            Double2DSpringSim.en.Y_VELOCITY + '-1',
            Double2DSpringSim.en.X_VELOCITY + '-2',
            Double2DSpringSim.en.Y_VELOCITY + '-2',
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME,
            Double2DSpringSim.en.ANCHOR_X,
            Double2DSpringSim.en.ANCHOR_Y
        ];
        const i18n_names = [
            Double2DSpringSim.i18n.X_POSITION + '-1',
            Double2DSpringSim.i18n.Y_POSITION + '-1',
            Double2DSpringSim.i18n.X_POSITION + '-2',
            Double2DSpringSim.i18n.Y_POSITION + '-2',
            Double2DSpringSim.i18n.X_VELOCITY + '-1',
            Double2DSpringSim.i18n.Y_VELOCITY + '-1',
            Double2DSpringSim.i18n.X_VELOCITY + '-2',
            Double2DSpringSim.i18n.Y_VELOCITY + '-2',
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME,
            Double2DSpringSim.i18n.ANCHOR_X,
            Double2DSpringSim.i18n.ANCHOR_Y
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.topMass_.setPosition(new Vector(0, 2));
        this.bob1_.setMass(0.5);
        this.bob2_.setMass(0.5);
        this.getSimList().add(this.topMass_, this.bob1_, this.bob2_, this.spring1_, this.spring2_);
        let pn;
        this.addParameter(new ParameterNumber(this, Double2DSpringSim.en.GRAVITY, Double2DSpringSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, Double2DSpringSim.en.DAMPING, Double2DSpringSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, Double2DSpringSim.en.LENGTH, Double2DSpringSim.i18n.LENGTH, () => this.getLength(), a => this.setLength(a)));
        this.addParameter(new ParameterNumber(this, Double2DSpringSim.en.MASS1, Double2DSpringSim.i18n.MASS1, () => this.getMass1(), a => this.setMass1(a)));
        this.addParameter(new ParameterNumber(this, Double2DSpringSim.en.MASS2, Double2DSpringSim.i18n.MASS2, () => this.getMass2(), a => this.setMass2(a)));
        this.addParameter(new ParameterNumber(this, Double2DSpringSim.en.STIFFNESS, Double2DSpringSim.i18n.STIFFNESS, () => this.getStiffness(), a => this.setStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
        this.restState();
        const vars = this.getVarsList().getValues();
        vars[U1X] += 0.5;
        vars[U1Y] -= 0.5;
        this.getVarsList().setValues(vars);
        this.saveInitialState();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', damping_: ' + Util.NF(this.damping_)
            + ', bob1_: ' + this.bob1_
            + ', bob2_: ' + this.bob2_
            + ', spring1_: ' + this.spring1_
            + ', spring2_: ' + this.spring2_
            + ', topMass_: ' + this.topMass_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'Double2DSpringSim';
    }
    ;
    restState() {
        const m1 = this.bob1_.getMass();
        const m2 = this.bob2_.getMass();
        const k1 = this.spring1_.getStiffness();
        const k2 = this.spring2_.getStiffness();
        const r1 = this.spring1_.getRestLength();
        const r2 = this.spring2_.getRestLength();
        const fixY = this.topMass_.getPosition().getY();
        const vars = this.getVarsList().getValues();
        vars[FIX_Y] = fixY;
        vars[U1X] = vars[U2X] = vars[FIX_X] = this.topMass_.getPosition().getX();
        vars[U1Y] = fixY - this.gravity_ * (m1 + m2) / k1 - r1;
        vars[U2Y] = fixY - this.gravity_ * (m2 / k2 + (m1 + m2) / k1) - r1 - r2;
        vars[V1X] = vars[V1Y] = vars[V2X] = vars[V2Y] = 0;
        this.getVarsList().setValues(vars);
        this.moveObjects(vars);
        this.potentialOffset_ = 0;
        this.setPEOffset(-this.getEnergyInfo().getPotential());
    }
    ;
    getEnergyInfo() {
        const ke = this.bob1_.getKineticEnergy() + this.bob2_.getKineticEnergy();
        let pe = this.gravity_ * this.bob1_.getMass() * this.bob1_.getPosition().getY();
        pe += this.gravity_ * this.bob2_.getMass() * this.bob2_.getPosition().getY();
        pe += this.spring1_.getPotentialEnergy();
        pe += this.spring2_.getPotentialEnergy();
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        this.bob1_.setPosition(new Vector(vars[U1X], vars[U1Y]));
        this.bob1_.setVelocity(new Vector(vars[V1X], vars[V1Y]));
        this.bob2_.setPosition(new Vector(vars[U2X], vars[U2Y]));
        this.bob2_.setVelocity(new Vector(vars[V2X], vars[V2Y]));
        this.topMass_.setPosition(new Vector(vars[FIX_X], vars[FIX_Y]));
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        this.dragBlock_ = -1;
        if (simObject == this.bob1_) {
            this.dragBlock_ = 0;
            return true;
        }
        else if (simObject == this.bob2_) {
            this.dragBlock_ = 1;
            return true;
        }
        else if (simObject == this.topMass_) {
            return true;
        }
        return false;
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const p = location.subtract(offset);
        if (simObject == this.topMass_) {
            va.setValue(FIX_X, p.getX());
            va.setValue(FIX_Y, p.getY());
        }
        else if (this.dragBlock_ == 0 || this.dragBlock_ == 1) {
            const block = this.dragBlock_ == 0 ? this.bob1_ : this.bob2_;
            if (simObject != block) {
                return;
            }
            const idx = 2 * this.dragBlock_;
            va.setValue(U1X + idx, p.getX());
            va.setValue(U1Y + idx, p.getY());
            va.setValue(V1X + idx, 0);
            va.setValue(V1Y + idx, 0);
            va.incrSequence(KE, PE, TE);
        }
        this.moveObjects(va.getValues());
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragBlock_ = -1;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        this.moveObjects(vars);
        change[TIME] = 1;
        const forces1 = this.spring1_.calculateForces();
        const f12 = forces1[1].getVector();
        const forces2 = this.spring2_.calculateForces();
        const f21 = forces2[0].getVector();
        const f22 = forces2[1].getVector();
        const m1 = this.bob1_.getMass();
        const m2 = this.bob2_.getMass();
        const b = this.damping_;
        if (this.dragBlock_ != 0) {
            change[U1X] = vars[V1X];
            change[U1Y] = vars[V1Y];
            change[V1X] = (f12.getX() + f21.getX() - b * vars[V1X]) / m1;
            change[V1Y] = -this.gravity_ + (f12.getY() + f21.getY() - b * vars[V1Y]) / m1;
        }
        if (this.dragBlock_ != 1) {
            change[U2X] = vars[V2X];
            change[U2Y] = vars[V2Y];
            change[V2X] = (f22.getX() - b * vars[V2X]) / m2;
            change[V2Y] = -this.gravity_ + (f22.getY() - b * vars[V2Y]) / m2;
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(Double2DSpringSim.en.GRAVITY);
    }
    ;
    getMass1() {
        return this.bob1_.getMass();
    }
    ;
    setMass1(value) {
        this.bob1_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(Double2DSpringSim.en.MASS1);
    }
    ;
    getMass2() {
        return this.bob2_.getMass();
    }
    ;
    setMass2(value) {
        this.bob2_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(Double2DSpringSim.en.MASS2);
    }
    ;
    getLength() {
        return this.springs_[0].getRestLength();
    }
    ;
    setLength(value) {
        for (let i = 0; i < this.springs_.length; i++) {
            this.springs_[i].setRestLength(value);
        }
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(Double2DSpringSim.en.LENGTH);
    }
    ;
    getStiffness() {
        return this.springs_[0].getStiffness();
    }
    ;
    setStiffness(value) {
        for (let i = 0; i < this.springs_.length; i++) {
            this.springs_[i].setStiffness(value);
        }
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(Double2DSpringSim.en.STIFFNESS);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(Double2DSpringSim.en.DAMPING);
    }
    ;
}
Double2DSpringSim.en = {
    ANCHOR_X: 'anchor X',
    ANCHOR_Y: 'anchor Y',
    X_POSITION: 'X position',
    Y_POSITION: 'Y position',
    X_VELOCITY: 'X velocity',
    Y_VELOCITY: 'Y velocity',
    DAMPING: 'damping',
    GRAVITY: 'gravity',
    MASS1: 'mass-1',
    MASS2: 'mass-2',
    LENGTH: 'spring length',
    STIFFNESS: 'spring stiffness',
    REST_STATE: 'rest state'
};
Double2DSpringSim.de_strings = {
    ANCHOR_X: 'Anker X',
    ANCHOR_Y: 'Anker Y',
    X_POSITION: 'X Position',
    Y_POSITION: 'Y Position',
    X_VELOCITY: 'X Geschwindigkeit',
    Y_VELOCITY: 'Y Geschwindigkeit',
    DAMPING: 'Dämpfung',
    GRAVITY: 'Gravitation',
    MASS1: 'Masse-1',
    MASS2: 'Masse-2',
    LENGTH: 'Federlänge',
    STIFFNESS: 'Federsteifheit',
    REST_STATE: 'ruhe Zustand'
};
Double2DSpringSim.i18n = Util.LOCALE === 'de' ? Double2DSpringSim.de_strings : Double2DSpringSim.en;
Util.defineGlobal('sims$springs$Double2DSpringSim', Double2DSpringSim);
